# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import re
import types
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import TypeVar

import pydantic

from .domainname import DomainName
from .stringtype import StringType


T = TypeVar('T', bound='ResourceName')
M = TypeVar('M', bound=Any)


class ResourceName(StringType):
    Typed: ClassVar[type['TypedResourceName[Any]']]
    model: Any
    relname: str
    service: DomainName
    patterns = [re.compile(r'//.*')]

    @property
    def id(self) -> str:
        return str.split(self, '/')[-1]

    @property
    def kind(self) -> str:
        return str.split(self.relname, '/')[0]

    @property
    def scalar(self) -> str:
        return self.id

    @classmethod
    def null(cls: type[T]) -> T: # pragma: no cover
        """Return an instance that represents an unassigned
        resource name.
        """
        return cls('//cochise.io/_/null')

    @classmethod
    def typed(cls, model: type[M]) -> TypedResourceName[M]:
        new_class =  types.new_class(
            name=f'{model.__name__}ResourceName',
            bases=(cls,),
            kwds={'model': model}
        )
        return cast(TypedResourceName[M], new_class)

    def __new__(cls: type[T], object: str) -> T:
        self = super().__new__(cls, object) # type: ignore
        if not str.startswith(object, '//'):
            raise ValueError("A resource name must start with slashes.")
        service, _, relname = str.partition(object[2:], '/')
        if not relname:
            raise ValueError("a valid ResourceName contains a relative name.")
        adapter = pydantic.TypeAdapter(DomainName)
        self.relname = relname
        self.service = adapter.validate_python(service)
        return cast(T, self)

    def __init_subclass__(cls, model: type[Any]) -> None:
        super().__init_subclass__()
        cls.model = model

    @classmethod
    def validate(cls, v: str, _: Any = None) -> str:
        return cls(v)


class TypedResourceName(ResourceName, Generic[M], model=pydantic.BaseModel):
    model: type[M]


ResourceName.Typed = TypedResourceName


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
from typing import cast
from typing import Annotated
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Self
from typing import TypeVar

import pydantic

from canonical.lib.types import SerializableSet
from canonical.lib.utils import deephash
from .apimodel import APIModel
from .apimodelfield import APIModelField
from .meta import APIVersionedMeta
from .ownerreference import OwnerReference
from .uidreference import UIDReference


N = TypeVar('N', bound=str)


class ObjectMeta(APIModel, Generic[N], group='core', version='v1'):
    model_config = {'populate_by_name': True}
    server_fields: ClassVar[set[str]] = {
        'created',
        'deleted',
        'finalizers',
        'generation',
        'resource_version',
        'owner_references',
        'updated',
        'uid'
    }

    annotations: Annotated[dict[str, Any], APIModelField(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store "
            "and retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects."
        ),
        when={'create', 'update', 'store', 'view'}
    )]

    labels: Annotated[dict[str, str | None], APIModelField(
        default_factory=dict,
        title="Labels",
        description=(
            "Map of string keys and values that can be used to organize and "
            "categorize (scope and select) objects."
        ),
        when={'create', 'update', 'store', 'view'}
    )]

    tags: Annotated[SerializableSet[str], APIModelField(
        default_factory=set,
        description=(
            "An array of tags that may be used to classify an object "
            "if a label or annotation is not applicable."
        ),
        when={'create', 'update', 'store', 'view'}
    )]

    created: Annotated[datetime.datetime | None, APIModelField(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Created",
        description=(
            "`created` is a timestamp representing the server time "
            "when this object was created. It is not guaranteed to be set in "
            "happens-before order across separate operations. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        ),
        frozen=True,
        when={'store', 'view'}
    )]

    deleted: Annotated[datetime.datetime | None, APIModelField(
        default=None,
        title="Deleted",
        description=(
            "The `deleted` field is an RFC 3339 date and time "
            "at which this resource will be deleted.\n\nThis field "
            "is set by the server when a graceful deletion is "
            "requested by the user, and is not directly settable "
            "by a client.The resource is expected to be deleted "
            "(no longer visible from resource lists, and not "
            "reachable by name) after the time in this field.\n\n"
            "Once set, this value may not be unset or be set "
            "further into the future, although it may be shortened "
            "or the resource may be deleted prior to this time."
        ),
        when={'store', 'view'}
    )]

    finalizers: Annotated[SerializableSet[str], APIModelField(
        default_factory=set,
        description=(
            "Must be empty before the object is deleted from the registry. "
            "Each entry is an identifier for the responsible component that "
            "will remove the entry from the list.\n\nIf the `deleted` field "
            "of the object is non-nil, entries in this list can only be "
            "removed. Finalizers may be processed and removed in any order."
        ),
        when={'store', 'view'}
    )]

    generation: Annotated[int, APIModelField(
        default=0,
        title="Generation",
        description=(
            "A sequence number representing a specific generation "
            "of the state. Populated by the system. Read-only."
        ),
        frozen=True,
        when={'store', 'view'}
    )]

    name: Annotated[N, APIModelField(
        default=...,
        title="Name",
        description=(
            "Name must be unique within a namespace. Is required when creating "
            "resources, although some resources may allow a client to request "
            "the generation of an appropriate name automatically. Name is "
            "primarily intended for creation idempotence and configuration "
            "definition. Cannot be updated."
        ),
        frozen=True,
        when={'create', 'store', 'view'}
    )]

    owner_references: Annotated[list[OwnerReference], APIModelField(
        default_factory=list,
        alias='ownerReferences',
        description=(
            "List of objects depended by this object. If ALL objects in the "
            "list have been deleted, this object will be garbage collected. "
            "\n\nIf this object is managed by a controller, then an entry in "
            "this list will point to this controller, with the controller "
            "field set to `true`. There cannot be more than one managing "
            "controller."
        ),
        frozen=True,
        when={'store', 'view'}
    )]

    resource_version: Annotated[str, APIModelField(
        default='',
        alias='resourceVersion',
        description=(
            "An opaque value that represents the internal version of this object "
            "that can be used by clients to determine when objects have changed. "
            "May be used for optimistic concurrency, change detection, and the "
            "watch operation on a resource or set of resources.\n\nClients must "
            "treat these values as opaque and passed unmodified back to the "
            "server. They may only be valid for a particular resource or set "
            "of resources. Populated by the system. Read-only. Value must be "
            "treated as opaque by clients and ."
        ),
        when={'store', 'view'}
    )]

    uid: Annotated[int, APIModelField(
        default=...,
        when={'store', 'view'},
        frozen=True
    )]

    updated: datetime.datetime = APIModelField(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Updated",
        description=(
            "`CreationTimestamp` is a timestamp representing the server time "
            "when this object was updated. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        ),
        frozen=True,
        when={'store', 'view'}
    )

    @classmethod
    def with_namespace(cls) -> type[Self]:
        args = cls.__pydantic_generic_metadata__.get('args')
        if not args:
            raise TypeError(
                "Can only create namespaced implementations of ObjectMeta "
                "from concrete classes."
            )
        t = type(
            f'Namespaced{cls.__name__}',
            (cls,),
            {
                '__annotations__': {
                    'name': cls.model_fields['name'].annotation,
                    'namespace': str
                },
                '__builder__': None,
                '__create_model__': None,
                '__mode__': 'domain',
                '__parameters__': args,
                'model_config': cls.model_config,
            },
        )
        return cast(type[Self], t)

    @pydantic.model_validator(mode='before') # type: ignore
    def preprocess(cls, values: dict[str, Any] | Self):
        if isinstance(values, dict):
            now = created = datetime.datetime.now(datetime.timezone.utc)
            if 'created' in cls.model_fields:
                created = values.setdefault('created', now)
            if 'updated' in cls.model_fields:
                values.setdefault('updated', created)
        return values

    @pydantic.field_serializer('resource_version', when_used='always', check_fields=False)
    def serialize_resource_name(self, value: str, _: Any) -> str:
        assert self.parent is not None
        return deephash(
            obj={
                **self.parent.model_dump(exclude={'metadata'}, mode='json'),
                'metadata': self.model_dump(exclude={'resource_version'}, mode='json')
            },
            encode='hex'
        )

    @property
    def key(self) -> UIDReference:
        if self.parent is None:
            raise TypeError(
                f"{type(self).__name__} is not attached to any known "
                "instance"
            )
        return UIDReference(
            api_version='v1', # type: ignore
            kind='ObjectDescriptor',
            uid=self.uid,
        )

    def cache_key(self, meta: APIVersionedMeta[Any]):
        return meta.cache_key(self.name, getattr(self, 'namespace', None))

    def dao(self):
        if self.parent is None:
            raise TypeError(
                f"{type(self).__name__} is not attached to any known "
                "instance"
            )
        assert self.parent is not None
        obj = self.__stored_model__.model_validate({
            **self.model_dump(),
            **self.parent.model_dump(include={'api_version', 'kind'}),
        })
        obj.add_to_parent(self.parent)
        return obj

    def get_controller(self) -> OwnerReference | None:
        """Return a :class:`OwnerReference` pointing to the
        controller, or ``None``.
        """
        ctrl: list[OwnerReference | None] = [
            o for o in self.owner_references if o.controller
        ]
        if len(ctrl) > 1:
            raise ValueError("Object has multiple controllers")
        ctrl.append(None)
        return ctrl.pop(0)

    def get_namespace(self):
        return getattr(self, 'namespace', None)

    def in_namespace(self, namespace: str):
        return getattr(self, 'namespace', None) == namespace


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.ext.api import ClusterObjectReference
from canonical.ext.oauth.models import ServerMetadata
from canonical.ext.oauth.clusterprovider import ClusterProvider
from canonical.ext.oauth.clientspec import ClientSpec
from canonical.ext.oauth.resources import AuthorizationStateBuilder
from .authorize import AuthorizeEndpoint


class UpstreamAuthorizationEndpoint(AuthorizeEndpoint):
    #: Metadata describing the upstream authorzation server.
    metadata: ServerMetadata

    async def discover(self, client: ClientSpec[ClusterObjectReference]):
        provider = await self.resources.get(
            client.provider.with_model(ClusterProvider),
        )
        meta = await provider.discover(self.http)
        self.metadata = meta

    async def get(
        self,
        builder: AuthorizationStateBuilder = fastapi.Depends(AuthorizationStateBuilder)
    ) -> fastapi.Response:
        assert self.metadata.authorization_endpoint is not None
        assert self.metadata.token_endpoint is not None
        builder.with_authorization_endpoint(self.metadata.authorization_endpoint)
        builder.with_issuer(self.metadata.issuer)
        builder.with_token_endpoint(self.metadata.token_endpoint)
        builder.own(self.client_resource, controller=True)
        builder.proxy(str(self.request.url_for('oauth2.callback')))
        builder.with_client(self.client)
        builder.with_params(self.params)
        if not self.params.redirect_uri:
            builder.with_redirect_uri(self.client.default_redirect())
        state = await builder
        await self.resources.persist(type(state), state)
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=state.get_user_agent_redirect_uri(self.params)
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical.ext.oauth.types import TokenTypeLiteral
from .grantmodel import GrantModel


class TokenExchangeRequest(GrantModel[Literal['urn:ietf:params:oauth:grant-type:token-exchange']]):
    audience: str | None = pydantic.Field(
        default=None,
        title="Audience",
        description=(
            "The logical name of the target service where the client intends "
            "to use the requested security token. This serves a purpose similar "
            "to the `resource` parameter but with the client providing a "
            "logical name for the target service. An OAuth client identifier, "
            "a SAML entity identifier, and an OpenID Connect Issuer Identifier "
            "are examples of things that might be used as audience parameter "
            "values."
        )
    )

    requested_token_type: TokenTypeLiteral | None = pydantic.Field(
        default=None,
        title="Requested Token Type",
        description=(
            "An identifier for the type of the requested security token. If "
            "the requested type is unspecified, the issued token type is at "
            "the discretion of the authorization server and may be dictated "
            "by knowledge of the requirements of the service or resource "
            "indicated by the resource or audience parameter."
        )
    )

    subject_token: str = pydantic.Field(
        default=...,
        title="Subject Token",
        description=(
            "A security token that represents the identity of the party on "
            "behalf of whom the request is being made. Typically, the subject "
            "of this token will be the subject of the security token issued "
            "in response to the request."
        )
    )

    subject_token_type: TokenTypeLiteral = pydantic.Field(
        default=...,
        title="Subject Token Type",
        description=(
            "An identifier, that indicates the type of the security token "
            "in the `subject_token` parameter."
        )
    )

    actor_token: str | None = pydantic.Field(
        default=None,
        title="Actor Token",
        description=(
            "A security token that represents the identity of the acting party. "
            "Typically, this will be the party that is authorized to use the "
            "requested security token and act on behalf of the subject."
        )
    )

    actor_token_type: TokenTypeLiteral | None = pydantic.Field(
        default=None,
        title="Actor Token Type",
        description=(
            "An identifier, that indicates the type of the security token in the "
            "`actor_token parameter`. This is **required** when the `actor_token` "
            "parameter is present in the request but **must not** be included "
            "otherwise."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.oauth.bases import AuthorizationServerEndpoint
from canonical.ext.oauth.models import ServerMetadata
from canonical.ext.oauth.protocols import IRegisteredClient


class MetadataEndpoint(AuthorizationServerEndpoint):
    name = 'oauth2.metadaya'
    path = '/.well-known/oauth-authorization-server'
    response_model = ServerMetadata
    summary = 'Metadata Endpoint'

    async def get(self) -> ServerMetadata:
        return ServerMetadata.model_validate({
            'issuer': f'{self.request.url.scheme}://{self.request.url.netloc}',
            'authorization_endpoint': str(self.request.url_for('oauth2.authorize')),
            'token_endpoint': str(self.request.url_for('oauth2.token')),
            'authorization_response_iss_parameter_supported': True
        })

    async def get_client(self) -> IRegisteredClient | None:
        return None


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import operator
from collections import abc
from collections import OrderedDict
from functools import reduce
from itertools import chain
from typing import Any
from typing import Callable

from ._deephash import deephash
from .deferred import Deferred
from .loader import import_symbol


__all__: list[str] = [
    'deephash',
    'import_symbol',
    'Deferred',
]


class class_property:
    __module__: str = 'oauthx.utils'

    def __init__(self, func: Callable[..., Any]):
        self.func = func

    def __get__(self, instance: Any, cls: Any) -> Any:
        return self.func(cls)


def merge_signatures(
    signatures: list[inspect.Signature],
    extra: list[inspect.Parameter] | None = None
) -> inspect.Signature:
    """Merge signatures to that FastAPI can inject the dependencies."""
    extra = extra or []
    params: dict[str, inspect.Parameter] = OrderedDict()
    for param in chain(chain(*[x.parameters.values() for x in signatures]), extra):
        if param.name in {'self', 'cls'}:
            continue
        if param.name.startswith('_'):
            continue
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            continue
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        params[param.name] = param

    return signatures[0].replace(
        parameters=list(sorted(params.values(), key=lambda p: (p.kind, p.default != inspect._empty))) # type: ignore
    )


def merged_call(func: Callable[..., Any], kwargs: Any) -> Any:
    sig = inspect.signature(func)
    return func(**{k: v for k, v in kwargs.items() if k in sig.parameters})


def throw(cls: type[Exception], *args: Any, **kwargs: Any):
    raise cls(*args, **kwargs)


def jsonpath(mapping: abc.Mapping[str, Any], path: str) -> Any:
    if not str.startswith(path, '/'):
        raise ValueError("Path must start with a slash.")
    parts: list[str] = []
    for p in str.split(str.lstrip(path, '/'), '/'):
        p = p.replace('~0', '~')
        p = p.replace('~1', '/')
        parts.append(p)
    return reduce(operator.getitem, parts, mapping)
    


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.ext.api import APIResourceModel
from canonical.ext.httpx import AsyncClient
from .providerspec import BaseProviderSpec


T = TypeVar('T', bound=BaseProviderSpec)


class BaseProvider(APIResourceModel[str], Generic[T], abstract=True):
    spec: T = pydantic.Field(
        default=...,
        description=(
            "Specification of the OAuth 2.x/OpenID Connect provider."
        )
    )

    def is_discoverable(self):
        return self.spec.is_discoverable()

    async def discover(self, client: AsyncClient):
        return await self.spec.discover(client)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import Literal
from typing import TypeVar
import pydantic


__all__: list[str] = [
    'ObjectReferenceModel'
]

T = TypeVar('T')


class ObjectReferenceModel(pydantic.BaseModel, Generic[T]):
    """Base class for object references."""
    model_config = {'populate_by_name': True}

    api_group: str = pydantic.Field(
        default=None,
        alias='apiGroup',
        title="API group",
        description=(
            "Specifies the API group of the referent. Cannot be updated."
        ),
        frozen=True
    )

    kind: str = pydantic.Field(
        default=...,
        description=(
            "Kind of the referent. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )

    name: T = pydantic.Field(
        default=...,
        description=(
            "The `.metadata.name` of of the referent. Cannot be "
            "updated."
        ),
        frozen=True
    )

    def __init_subclass__(
        cls,
        api_group: str | None = None,
        kind: str | None = None,
        **kwargs: Any
    ) -> None:
        super().__init_subclass__()
        fields = cls.model_fields
        if api_group is not None:
            fields['api_group'].default = api_group
            fields['api_group'].annotation = Literal[f'{api_group}'] # type: ignore
        if kind is not None:
            fields['kind'].default = kind
            fields['kind'].annotation = Literal[f'{kind}'] # type: ignore
        cls.model_rebuild()
        cls.initialize_class(**kwargs)

    @classmethod
    def initialize_class(cls, **kwargs: Any):
        pass

    def is_local(self) -> bool:
        """Return a boolean indicating if the reference is to a local
        object i.e. on the same server. The default returns ``False``
        so that any local reference has to be explicitely implemented..
        """
        return False

    def get_namespace(self) -> str | None:
        return None

    def is_namespaced(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from collections import OrderedDict
from inspect import Parameter
from inspect import Signature
from typing import Any
from typing import TypeVar

import fastapi

from canonical.ext.resource import Error
from canonical.ext.resource import ResourceType
from ..response import Response
from .detail import Detail


T = TypeVar('T')


class Replace(Detail[T]):
    detail = True
    existing = True
    method = 'PUT'
    requires_body = True
    verb = 'replace'

    def annotate_handler(self, signature: Signature) -> Signature:
        params: OrderedDict[str, Parameter] = OrderedDict(signature.parameters)
        for name, param in signature.parameters.items():
            if param.annotation == self.model:
                params[name] = param.replace(annotation=self.get_input_model())
                self.input_params.add(name)
                continue
        return signature.replace(parameters=list(params.values()))

    def can_replace(self, resource: ResourceType) -> bool:
        return resource.replacable()

    def get_endpoint_summary(self) -> str:
        return f'Replace an existing {self.model.__name__}'

    def get_openapi_responses(self) -> dict[int, Any]:
        return {
            **super().get_openapi_responses(),
            409: {
                'model': Error,
                'description': (
                    f'The {self.model.__name__} identified by the path parameters '
                    'can not be replaced.'
                )
            }
        }

    def get_response_description(self) -> str:
        return f"The latest version of the {self.model.__name__} object."

    def replaces(self) -> bool:
        return True

    async def render_to_response(
        self,
        request: fastapi.Request,
        result: Any,
        media_type: str
    ) -> Response[Any]:
        if not isinstance(result, self.model):
            raise TypeError(
                f"Function handle() must return a {self.model.__name__} "
                f"instance, got {type(result).__name__}."
            )
        return Response(
            media_type=media_type,
            status_code=205,
            content=result
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Self

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler

from canonical.ext.api import get_meta
from canonical.ext.api import APIResourceType
from canonical.ext.api.protocols import IObjectIdentifier
from canonical.lib.utils.encoding import b64decode
from canonical.lib.utils.encoding import b64encode


class URLSafeClientID:
    supported_types: ClassVar[set[str]] = {
        'clusterclients'
    }
    version: str
    plural: str
    namespace: str
    name: str

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.is_instance_schema(str),
                    core_schema.no_info_plain_validator_function(cls.fromstring),
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.encode)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def encode(cls, self: Self | str):
        if isinstance(self, str):
            return self
        return b64encode(f'{self.version}:{self.plural}:{self.namespace}:{self.name}', encoder=str)

    @classmethod
    def fromstring(cls, v: str):
        try:
            version, plural, namespace, name = str.split(b64decode(v, str), ':')
            if version != '1':
                raise ValueError
            if plural not in cls.supported_types:
                raise ValueError
            return cls(version, plural, namespace, name)
        except ValueError:
            raise ValueError("Invalid Client ID.")

    @property
    def cache_key(self) -> str:
        return self.meta.cache_key(self.name, self.namespace or None)

    @property
    def meta(self):
        return get_meta('oauth/v2', plural=self.plural)

    @property
    def model(self):
        return self.meta.model

    @property
    def ref(self) -> IObjectIdentifier[APIResourceType]:
        return cast(
            IObjectIdentifier[APIResourceType],
            self.meta.reference(self.name, namespace=self.namespace or None)
        )

    def __init__(self, version: str, plural: str, namespace: str, name: str):
        self.version = version
        self.plural = plural
        self.namespace = namespace
        self.name = name

    def __repr__(self):
        return (
            f"<URLSafeClientID({repr(self.version)}, {repr(self.plural)}, "
            f"{repr(self.namespace)}, {repr(self.name)})"
        )

    def __str__(self) -> str:
        return self.encode(self)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from typing import cast
from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Any
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client

from canonical.protocols import ITransaction
from canonical.ext.api import APIVersionedMeta
from canonical.ext.api import APIVersioned
from canonical.ext.api import ClusterObjectReference
from canonical.ext.api import DefaultInspector
from canonical.ext.api import LocalObjectReference
from canonical.ext.api import ObjectReference
from canonical.ext.api import OwnerReference
from canonical.ext.api import UIDReference
from canonical.ext.api.bases import BaseReference
from canonical.ext.api.annotations import Reference
from canonical.ext.api.annotations import APIResourceType
from canonical.ext.api.protocols import IObjectIdentifier
from canonical.ext.resource import BaseResourceRepository
from .basedatastorestorage import BaseDatastoreStorage
from .protocols import IDatastoreKey
from .protocols import IDatastoreEntity


B = TypeVar('B')
T = TypeVar('T', bound=APIResourceType)
SupportedKeyTypes = ClusterObjectReference, LocalObjectReference, ObjectReference, OwnerReference, UIDReference
KeyType = ClusterObjectReference | LocalObjectReference | ObjectReference | OwnerReference | UIDReference


class GoogleDatastoreResourceRepository(BaseResourceRepository):
    backend: BaseDatastoreStorage
    inspector = DefaultInspector
    logger: logging.Logger = logging.getLogger('uvicorn')

    def __init__(
        self,
        backend: BaseDatastoreStorage,
        **kwargs: Any
    ) -> None:
        super().__init__(inspector=self.inspector, **kwargs)
        self.backend = backend

    def all(self, model: type[B], namespace: str | None = None) -> BaseResourceRepository.Cursor[B]:
        return self.query(model=model, namespace=namespace)

    def get_entity_name(self, key: KeyType | type[T] | APIVersionedMeta[Any]) -> str:
        if isinstance(key, UIDReference):
            name = f'{key.api_group}/{key.kind}'
        elif isinstance(key, BaseReference):
            key = self.inspector.inspect(cast(type[APIVersioned], key.get_model()))
        else:
            key = self.inspector.inspect(cast(type[APIVersioned], key))
        name = f'{key.api_group}/{key.kind}'
        if not key.api_group:
            name = key.kind
        return name

    def model_factory(
        self,
        key: IDatastoreKey,
        entity: Mapping[str, Any],
        model: type[T]
    ) -> T:
        return model.model_validate(dict(entity))

    def resource_key(
        self,
        key: IObjectIdentifier[T] | Reference,
        parent: IDatastoreKey |None = None,
        namespace: str | None = None
    ) -> IDatastoreKey:
        assert isinstance(key, SupportedKeyTypes)
        return self.backend.entity_key(
            self.get_entity_name(key),
            key.as_name(),
            parent=parent,
            namespace=namespace or key.get_namespace()
        )

    def query(
        self,
        model: type[B],
        filters: Iterable[tuple[str, str, Any]] | None = None,
        sort: Iterable[str] | None = None,
        namespace: str | None = None,
        limit: int | None = None,
        kind: str | None = None,
        page_size: int = 10,
        keys: list[Any] | None = None,
        **_: Any
    ) -> BaseResourceRepository.Cursor[B]:
        q = self.backend.query(
            model=cast(type[pydantic.BaseModel], model),
            filters=filters,
            sort=sort,
            namespace=namespace,
            limit=limit,
            kind=kind or self.get_entity_name(cast(type[APIResourceType], model)),
            page_size=page_size,
            keys=keys
        )
        return cast(BaseResourceRepository.Cursor[B], q)
        

    async def allocate(self, obj: type[APIResourceType]) -> int:
        k = self.get_entity_name(obj)
        i = await self.backend.allocate_identifier(k)
        self.logger.info("Allocated identifier (kind: %s, uid: %s)", k, i)
        return i

    async def exists(self, key: Reference | Iterable[tuple[str, str, Any]], model: type[T]) -> bool:
        assert isinstance(key, BaseReference), type(key)
        q = self.backend.query(
            model=model,
            namespace=key.get_namespace(),
            keys=[self.resource_key(key, namespace=key.get_namespace())],
            page_size=1
        )
        return await q.exists()

    async def get_entity_by_key(self, key: IDatastoreKey) -> IDatastoreEntity | None:
        return await self.backend.get_entity_by_key(key)

    async def reference(
        self,
        key: IObjectIdentifier[T] | Reference,
        model: type[T],
    ) -> T | None:
        k = self.resource_key(key, namespace=key.get_namespace())
        e = await self.get_entity_by_key(key=k)
        instance = None
        if e is not None:
            instance = await self.restore(self.model_factory(k, dict(e), model))
        return instance

    def transaction(self, transaction: Any | None = None):
        return self.backend.transaction(transaction)

    async def persist_entity(
        self,
        client: Client | ITransaction,
        entity: IDatastoreEntity
    ) -> IDatastoreEntity:
        return await self.backend.run_in_executor(functools.partial(client.put, entity)) # type: ignore

    async def put(
        self,
        key: IObjectIdentifier[T] | Reference,
        dao: pydantic.BaseModel,
        transaction: Any | None = None
    ) -> None:
        await self.persist_entity(
            client=transaction or self.backend.client,
            entity=self.backend.entity_factory(
                key=self.resource_key(key),
                obj=dao
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

import httpx
import pydantic

from canonical.ext.api import APIModelField
from canonical.ext.api import ResourceSpec
from canonical.ext.crypto import EncryptionResult
from canonical.ext.oauth.protocols import ITokenRequest
from .client import Client
from .models.responses import TokenResponse
from .types import AuthenticationMethod
from .types import Error
from .types import ProtocolViolation
from .types import RedirectURI
from .types import ResponseTypeLiteral


RefType = TypeVar('RefType')


class ClientSpec(ResourceSpec, Generic[RefType]):
    client_id: str = pydantic.Field(
        default=...,
        alias='clientId',
        title='Client ID',
        description=(
            "The OAuth 2.x/OpenID Connect client identifier."
        ),
        max_length=64
    )

    client_secret: EncryptionResult | str = APIModelField(
        default=...,
        alias='clientSecret',
        title='Client secret',
        description=(
            "The OAuth 2.x/OpenID Connect client secret."
        ),
        max_length=64,
        encrypt=True
    )

    provider: RefType = pydantic.Field(
        default=...,
        description=(
            "A local reference to an OAuth 2.x/OpenID Connect provider."
        )
    )

    response_type: ResponseTypeLiteral = pydantic.Field(
        default='code',
        title="Response type",
        alias='responseType',
        description=(
            "The default value for the `response_type` parameter in an "
            "authorization request."
        )
    )

    internal_redirect_uris: list[RedirectURI] = pydantic.Field(
        default_factory=list,
        title="Allowed redirection URIs",
        alias='internalRedirectUris',
        description=(
            "Specifies the set of URIs to which the server may "
            "redirect after an authorization request with the "
            "upstream authorization server."
        ),
    )

    authentication_method: AuthenticationMethod = pydantic.Field(
        default='client_secret_post',
        title="Default Authentication Method",
        description=(
            "Specifies the default authentication method that the client "
            "will use for all endpoints, unless the server is discoverable "
            "and its metadata tells the client to use a different method."
        )
    )

    @pydantic.field_validator('internal_redirect_uris', mode='after')
    def validate_redirect_uris(cls, value: list[RedirectURI]):
        if len(value) == 0:
            raise ValueError("At least one redirect uri must be specified.")
        return value

    @property
    def id(self) -> str:
        return self.client_id

    def allows_redirect(self, uri: str | None) -> bool:
        return uri is None or any([
            str(x) == uri for x
            in self.internal_redirect_uris
        ])

    def allows_response_type(self, response_type: str) -> bool:
        return self.response_type == response_type

    def configure(self, **kwargs: Any):
        kwargs.update({
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'mode': self.authentication_method
        })
        return Client(**kwargs)

    def default_redirect(self) -> str:
        return str(self.internal_redirect_uris[0])

    def must_encrypt(self, attname: str) -> bool:
        return attname == 'client_secret'

    async def authenticate(self, request: ITokenRequest) -> None:
        if self.authentication_method != 'client_secret_post':
            raise NotImplementedError(self.authentication_method)
        assert not isinstance(self.client_secret, EncryptionResult)
        request.identify(self.client_id)
        request.set_client_secret(self.authentication_method, self.client_secret)

    async def obtain(
        self,
        grant: Any,
        http: httpx.AsyncClient | None = None
    ) -> TokenResponse | Error | ProtocolViolation:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dependency import Dependency
    from .requirement import Requirement


class DependencyProvider:
    dependencies: dict[str, 'Dependency']
    requirements: dict[str, 'Requirement']

    def __init__(self):
        self.requirements = {}
        self.dependencies = {}

    def inject(self):
        for requirement in self.requirements.values():
            requirement.inject(self.dependencies[requirement.name])

    def require(self, requirement: Requirement):
        self.requirements[requirement.name] = requirement

    def provide(self, dependency: Dependency):
        self.dependencies[dependency.name] = dependency


__provider = DependencyProvider()
inject = __provider.inject
require = __provider.require
provide = __provider.provide


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.lib.types import UnixTimestamp

A = TypeVar('A')
K = TypeVar('K')
O = TypeVar('O')
U = TypeVar('U')


class BaseJWK(pydantic.BaseModel, Generic[K, A, O, U]):
    alg: A | None = None
    crv: str | None = None
    exp: UnixTimestamp | None = None
    iat: UnixTimestamp | None = None
    key_ops: list[O] | None = None
    kid: str | None = None
    kty: K
    nbf: UnixTimestamp | None = None
    use: U | None = None

    @property
    def thumbprint(self) -> str:
        raise NotImplementedError

    def verify(
        self,
        signature: bytes,
        payload: bytes,
        *,
        alg: str | None = None
    ) -> bool:
        raise NotImplementedError


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import operator
from functools import reduce
from typing import cast
from typing import Annotated
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Protocol
from typing import Self
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic
from pydantic.fields import FieldInfo

from canonical.exceptions import ProgrammingError
from .apimodel import APIModel
from .conditiontype import ConditionType
from .objectmeta import ObjectMeta
if TYPE_CHECKING:
    from .statefulresource import StatefulResource


C = TypeVar('C', bound=ConditionType[Any])
S = TypeVar('S', contravariant=True)
RESERVED_FIELDS: set[str] = {
    'name',
    'namespace',
    'uid'
}

ConditionList = Annotated[
    list[ConditionType[Any]],
    pydantic.Field(
        description=(
            "Conditions describe specific events related to this {kind}."
        ),
    )
]


class IHandler(Protocol, Generic[S]):
    def __call__(self, status: S, *args: Any) -> None: ...


class BaseResourceStatus(APIModel):
    _adapter: ClassVar[pydantic.TypeAdapter[Any]]
    _storage: ClassVar[pydantic.TypeAdapter[Any]]

    changed: datetime.datetime = pydantic.Field(
        default=...,
        description=(
            "The date and time of the last change to this {kind}."
        )
    )

    version: int = pydantic.Field(
        default=...,
        description=(
            "Current version of the {kind}."
        )
    )

    message: str = pydantic.Field(
        default=...,
        description="The message of the last known condition."
    )

    current: str = pydantic.Field(
        default=...,
        description="The status of the last known condition."
    )


class ResourceStatus(BaseResourceStatus):
    __abstract__: ClassVar[bool] = True
    _handlers: ClassVar[dict[str, tuple[ConditionType[Self], IHandler[Self]]]] = {}
    ClusterModel: ClassVar[type[ConditionType[Self]]]

    _resource: 'StatefulResource[Any, Any]' = pydantic.PrivateAttr(
        default=None
    )

    _dirty: bool = pydantic.PrivateAttr(
        default=False,
    )

    conditions: ConditionList

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def namespace(self) -> str:
        raise NotImplementedError

    @property
    def uid(self) -> int | str:
        raise NotImplementedError

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        super().__pydantic_init_subclass__(**kwargs)
        reserved = set(cls.model_fields) & RESERVED_FIELDS
        if reserved:
            raise ProgrammingError(
                f"{cls.__name__} contains reserved fields: "
                f"{sorted(reserved)}"
            )
        if cls.__abstract__:
            cls.__abstract__ = False

        # Create a model for the conditions that includes
        # the UID of the Resource. This allows persistencee
        # in a different storage container (e.g. table).
        cls._handlers = {}
        cls._cluster_model = type( # type: ignore
            cls.__name__,
            (BaseResourceStatus,),
            {
                '__cluster__': True,
                '__annotations__': {
                    'namespace': str | None,
                    'name': str,
                    'uid': int | str
                }
            }
        )

    @classmethod
    def contribute_to_class(cls, parent: Any, attname: str, field: FieldInfo) -> None:
        for field in cls.model_fields.values():
            if field.description is None:
                continue
            field.description = str.format(field.description, kind=parent.__name__)
        assert cls.model_rebuild(force=True)

    @classmethod
    def default(cls) -> Self:
        raise NotImplementedError(
            f"{cls.__name__}.default() is not implemented. Subclasses "
            "must override this method to initialize the default "
            "state."
        )

    @classmethod
    def factory(cls, status: str, message: str | None = None, **kwargs: Any):
        now = datetime.datetime.now(datetime.timezone.utc)
        kwargs.setdefault('changed', now)
        kwargs.setdefault('conditions', [])
        kwargs.setdefault('current', status)
        kwargs.setdefault('message', message or '')
        kwargs.setdefault('version', 1)
        return cls.model_validate(kwargs)

    @classmethod
    def on(cls, status: str, *, title: str | None = None, **annotations: Any):
        """Create a function decorator that adds the specified `status`
        to the model with the given list of attributes `annotations`.
        """

        def decorator(func: Callable[..., None]) -> Callable[..., None]:
            reserved = set(annotations) & {'replay', *set(ConditionType.model_fields)}
            if reserved:
                raise ProgrammingError(
                    f"Handler {func.__name__} declares reserved fields: "
                    f"{sorted(reserved)}"
                )
            model = type('Condition', (ConditionType,), {
                'handle': staticmethod(func),
                'model_config': {
                    'populate_by_name': True,
                    'title': title or str.title(status) 
                },
                '__annotations__': {
                    **annotations,
                    'status': Literal[f'{status}']
                },
                '__handler_args__': set(annotations)
            })
            cls._handlers[status] = (cast(type[ConditionType[Self]], model), func) # type: ignore

            # Rebuild the .conditions field to include the new model so
            # that it is properly parsed and documented.
            models = [h[0] for h in cls._handlers.values()]

            annotation = model
            if len(models) > 1:
                annotation = reduce(operator.or_, models)
            cls.model_fields['conditions'].annotation = list[annotation]

            # Create an adapter to deserialize and serialize the storage
            # models for the conditions.
            cls._storage = pydantic.TypeAdapter(annotation)

            # Reinitialize the adapter and rebuild the model so that all
            # fields and validators are up-to-date.
            cls._adapter = pydantic.TypeAdapter(annotation)
            assert cls.model_rebuild(force=True)
            return func

        return decorator

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any]):
        values.setdefault('conditions', [])
        return values

    @property
    def adapter(self) -> pydantic.TypeAdapter[ConditionType[Self]]:
        return self._adapter

    @property
    def resource(self) -> 'StatefulResource[ObjectMeta[Any], Any]':
        return self._resource

    def apply(self, status: str, **kwargs: Any):
        kwargs.setdefault('timestamp', datetime.datetime.now(datetime.timezone.utc))
        condition = self.adapter.validate_python({
            'status': status,
            'observed_generation': 1,
            #'observed_generation': self.resource.metadata.generation,
            **kwargs
        })
        self._apply(condition)

    def attach(self, resource: 'StatefulResource[Any, Any]'):
        self._resource = resource

    def clusterstate(self) -> Self:
        state = self._cluster_model.model_validate({
            **self.model_dump(exclude={'conditions'}),
            'name': self.resource.metadata.name,
            'namespace': self.resource.get_namespace(),
            'uid': self.resource.metadata.uid
        })
        return cast(Self, state)

    def is_dirty(self) -> bool:
        return self._dirty

    def is_final(self):
        return self.resource.is_final()

    def mark_dirty(self) -> None:
        self._dirty = True

    def model_dump_storage(self) -> dict[str, Any]:
        return self._storage.dump_python(self)

    def storage_model(self, data: Any) -> ConditionType[Self]:
        return self._storage.validate_python(data)

    def _apply(
        self,
        condition: ConditionType[Self],
        replay: bool = False
    ) -> ConditionType[Self]:
        if self.is_final() and not replay:
            raise ValueError("Resource is in its final state.")
        if condition.timestamp == self.changed:
            condition.timestamp += datetime.timedelta(microseconds=1)
        self.changed = condition.timestamp
        self.current = condition.status
        self.message = condition.message
        condition.observed_generation = self.resource.metadata.generation
        condition.apply(self)
        if not replay:
            self.conditions.append(condition)
        if not replay:
            self.mark_dirty()
        return condition


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any

import pydantic

from canonical.lib.types import Base64
from canonical.lib.utils.encoding import b64encode
from canonical.lib.utils.encoding import b64decode_json
from canonical.ext.jose.protocols import ITokenSignatureVerifier
from canonical.ext.jose.types import JWTDict
from canonical.ext.jose.types import JOSEHeaderDict
from .jwsheader import JWSHeader


class Signature(pydantic.BaseModel):
    claims: JWSHeader
    protected: bytes | None = None
    header: dict[str, Any] = {}
    signature: Base64

    @pydantic.field_validator('protected', mode='before')
    def preprocess_protected(cls, value: bytes | str | None) -> bytes | None:
        if isinstance(value, str):
            value = str.encode(value, 'ascii')
        return value
    
    @pydantic.model_validator(mode='before') # type: ignore
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        # The Header Parameter values used when creating or validating
        # individual signature or MAC values are the union of the two
        # sets of Header Parameter values that may be present: (1) the
        # JWS Protected Header represented in the "protected" member of
        # the signature/MAC's array element, and (2) the JWS Unprotected
        # Header in the "header" member of the signature/MAC's array element.
        # The union of these sets of Header Parameters comprises the JOSE
        # Header.  The Header Parameter names in the two locations MUST
        # be disjoint.
        claims = values.get('header') or {}
        protected = {}
        if values.get('protected'):
            protected = b64decode_json(values['protected'])
        if not isinstance(protected, dict):
            raise ValueError("The encoded protected header must be a JSON object.")
        if set(claims.keys()) & set(protected.keys()):
            raise ValueError(
                "The header parameter names in the protected and "
                "unprotected header must be disjoint."
            )
        values['claims'] = {**claims, **protected}
        return values

    @property
    def typ(self):
        return self.claims.typ

    def encode(self, payload: bytes) -> bytes:
        if self.protected is None: # pragma: no cover
            raise ValueError("Missing protected header.")
        return bytes.join(b'.', [
            self.protected,
            payload,
            b64encode(self.signature)
        ])

    async def verify(
        self,
        verifier: ITokenSignatureVerifier,
        claims: JWTDict,
        payload: bytes
    ) -> bool:
        assert self.protected is not None
        return await verifier.verify_token(
            header=cast(JOSEHeaderDict, self.claims.model_dump(by_alias=True)),
            claims=claims,
            signature=self.signature,
            payload=bytes.join(b'.', [self.protected, payload])
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


T = TypeVar('T')


class SerializableSet(set[T]):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.list_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.no_info_plain_validator_function(cls)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(list)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.list_schema())


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import TypeVar

import pydantic

from .resource import BaseResource


T = TypeVar('T')


class BaseRootResource(pydantic.RootModel[T], BaseResource):

    def replacable(self) -> bool:
        return cast(BaseResource, self.root).replacable()

    def replace(self, instance: pydantic.BaseModel):
        return cast(BaseResource, self.root).replace(instance)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .apiversioned import APIVersioned
from .basereference import BaseReference
from .fields import APIVersion
from .fields import Kind
from .fields import UID


__all__: list[str] = [
    'OwnerReference'
]


class OwnerReference(APIVersioned, BaseReference):
    """An `OwnerReference` instance contains enough information to let you
    identify an owning object. An owning object must be in the same namespace
    as the dependent, or be cluster-scoped, so there is no namespace field.
    """

    api_version: APIVersion = pydantic.Field(
        default=...,
        description="API version of the referent."
    )

    block_owner_deletion: bool = pydantic.Field(
        default=True,
        alias='blockOwnerDeletion',
        description=(
            "If true, AND if the owner has the `foregroundDeletion` finalizer, "
            "then the owner cannot be deleted from the key-value store until "
            "this reference is removed."
        )
    )

    controller: bool = pydantic.Field(
        default=False,
        description=(
            "If true, this reference points to the managing controller."
        )
    )

    kind: Kind = pydantic.Field(
        default=...,
        description="Kind of the referent."
    )

    name: str = pydantic.Field(
        default=...,
        description="Name of the referent."
    )

    uid: UID = pydantic.Field(
        default=...,
        description="UID of the referent"
    )

    def attach_to_namespace(self, namespace: str) -> None:
        pass

    def is_namespaced(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Iterable

import fastapi

from canonical.ext.api import APIResourceType
from canonical.ext.iam import BaseAuthorizationContext
from canonical.ext.iam import ClusterRole
from canonical.ext.iam import ClusterRoleBinding
from canonical.ext.iam import Role
from canonical.ext.iam import RoleBinding
from canonical.ext.iam.types import Permission
from canonical.ext.iam.types import PermissionSet
from canonical.ext.namespace import Namespace
from canonical.lib.protocols import ICache
from .annotations import DefaultInspector
from .resourcerepository import ResourceRepository


class LocalAuthorizationContext(BaseAuthorizationContext):
    cache_ttl: int = 900
    logger: logging.Logger = logging.getLogger('uvicorn')
    roles: set[str]
    _cluster_roles: dict[str, ClusterRole] = NotImplemented
    _ready: bool = False

    @property
    def cache(self) -> ICache:
        return getattr(self.request.state, 'cache')

    @property
    def email(self) -> str | None:
        return getattr(self.request.state, 'email', None)

    def __init__(
        self,
        inspector: DefaultInspector,
        repo: ResourceRepository,
        request: fastapi.Request,
    ):
        super().__init__()
        self.inspector = inspector
        self._cluster_permissions = LocalAuthorizationContext._cluster_roles
        self.granted = PermissionSet()
        self.repo = repo
        self.request = request
        self.subject_type = 'User'

    async def get_cluster_bindings(self):
        cache_key = f'clusterrolebindings:users:{self.email or "allUsers"}'
        bindings: list[ClusterRoleBinding] | None = await self.cache.get(
            key=cache_key,
            decoder=list[ClusterRoleBinding]
        )
        if bindings is None:
            bindings = []
            public = self.repo.query(
                model=ClusterRoleBinding,
                filters=[
                    ('subjects.kind', '=', self.subject_type),
                    ('subjects.name', '=', 'allUsers'),
                ]
            )
            if self.email is not None:
                self.logger.debug(
                    "Retrieving global role bindings (kind: %s, email: %s)",
                    self.subject_type,
                    self.email
                )
                user = self.repo.query(
                    model=ClusterRoleBinding,
                    filters=[
                        ('subjects.kind', '=', self.subject_type),
                        ('subjects.name', '=', self.email),
                    ]
                )
                bindings = [*await user]
            bindings = [*bindings, *await public]
            await self.cache.set(
                key=cache_key,
                value=bindings,
                encoder=list[ClusterRoleBinding],
                ttl=self.cache_ttl
            )
        return bindings

    def get_namespace_roles(self, roles: Iterable[str]):
        assert self.namespace is not None
        return self.repo.query(
            model=Role,
            # TODO: Current implementation causes the query
            # to return not any object if there are multiple
            # roles.
            #filters=[('metadata.name', '=', list(roles))],
            namespace=self.namespace
        )

    def is_authenticated(self) -> bool:
        return self.email is not None

    def is_authorized(self) -> bool:
        return self.is_granted(self.permission)

    def is_granted(self, permission: str) -> bool:
        return self.granted.has(permission)

    async def get_permissions(self, permissions: set[str]) -> PermissionSet:
        return PermissionSet({Permission(p) for p in permissions if self.is_granted(p)})

    async def get_namespace_permissions(self, roles: Iterable[str]) -> PermissionSet:
        permissions = PermissionSet()
        async for role in self.get_namespace_roles(roles):
            if role.metadata.name not in roles:
                continue
            permissions.update(role.permissions)
        return permissions

    async def has(self, permissions: str | set[str]) -> bool:
        if isinstance(permissions, str):
            permissions = {permissions}
        granted = await self.get_permissions(permissions)
        return set(granted) == permissions

    async def get_namespace_bindings(self):
        cache_key = f'rolebindings:users:{self.email}'
        bindings = await self.cache.get(cache_key, list[RoleBinding])
        if bindings is None:
            self.logger.debug(
                "Retrieving local role bindings (namespace: %s, kind: %s, email: %s)",
                self.namespace,
                self.subject_type,
                self.email
            )
            q = self.repo.query(
                model=RoleBinding,
                filters=[
                    ('subjects.kind', '=', self.subject_type),
                    ('subjects.name', '=', str(self.email)),
                ],
                namespace=self.namespace
            )
            bindings = [x async for x in q]
            await self.cache.set(cache_key, bindings, list[RoleBinding], ttl=self.cache_ttl)
        return bindings

    async def setup(self):
        self.model: type[APIResourceType] = getattr(self.request.state, 'model')
        self.meta = self.inspector.inspect(self.model)

        assert self.meta.plural
        self.namespace: str | None = getattr(self.request.state, 'namespace', None)
        self.verb: str = getattr(self.request.state, 'verb')
        self.api_group = self.meta.api_group
        self.plural = self.meta.plural
        self.permission = self.get_permission_name(self.api_group, self.plural, self.verb)
        self.logger.debug(
            "Endpoint permission is %s (endpoint: %s).",
            self.permission,
            self.request.url.path
        )

        # Namespace is a special cause because permissions in the namespace
        # also apply to the namespace itself.
        if self.namespace is None and self.model == Namespace:
            # This will be none for cluster resources.
            self.namespace = getattr(self.request.state, 'name', None)
    
        await self.setup_cluster()

        scoped_roles: set[str] = set()
        global_roles: set[str] = set()
        try:
            for binding in await self.get_cluster_bindings():
                role = self._cluster_roles[binding.role_ref.name]
                self.granted.update(role.permissions)
        except KeyError:
            # Occurs when cluster roles have been added and made the
            # cache stale.
            return
        if self.namespace is not None:
            for obj in await self.get_namespace_bindings():
                if obj.is_global():
                    global_roles.add(obj.role_ref.name)
                else:
                    scoped_roles.add(obj.role_ref.name)

            self.granted |= await self.get_namespace_permissions(scoped_roles)
        self.logger.debug(
            "Granted permissions are %s (endpoint: %s, subject: %s)",
            ', '.join(sorted(map(str, self.granted))),
            self.request.url,
            self.email
        )

    async def get_cluster_roles(self) -> list[ClusterRole]:
        meta = self.inspector.inspect(ClusterRole)
        cache_key = meta.cache_prefix
        roles = await self.cache.get(cache_key, list[ClusterRole])
        if roles is None:
            self.logger.debug("Retrieving cluster roles and permissions")
            roles = [x async for x in self.repo.all(ClusterRole)]
            await self.cache.set(cache_key, roles, list[ClusterRole], ttl=3600)
        return roles

    async def setup_cluster(self):
        if self._cluster_roles != NotImplemented:
            return
        self._cluster_roles = {
            x.metadata.name: x
            for x in await self.get_cluster_roles()
        }


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid
from typing import Any

import fastapi

from canonical.ext.templates import TemplateService


class UserAgentException(Exception):
    template_name: str
    page_heading: str | None = None
    page_title: str | None = None
    status_code: int = 400

    def get_template_names(self) -> list[str]:
        return [self.template_name]

    def get_template_context(self) -> dict[str, Any]:
        return {
            'exception': self,
            'incident_id': str(uuid.uuid4()),
            'page_heading': self.page_heading,
            'page_title': self.page_title
        }

    async def render(
        self,
        templates: TemplateService,
        context: dict[str, Any] | None = None
    ) -> fastapi.Response:
        return fastapi.Response(
            status_code=self.status_code,
            content=await templates.render_template(
                templates=self.get_template_names(),
                context={
                    **(context or {}),
                    **self.get_template_context()
                }
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from canonical.lib.types import StringType


M = TypeVar('M', bound=pydantic.BaseModel)


class JWSCompactEncoded(StringType):

    @classmethod
    def validate(cls, v: str):
        if not v.count('.') == 2:
            raise ValueError("Invalid JWS Compact Encoding.")
        return cls(v)

    def compact(self):
        return self

    def jose(self, factory: type[M]) -> M:
        protected, payload, signature = str.split(self, '.')
        return factory.model_validate({
            'payload': payload,
            'signatures': [{
                'protected': protected,
                'signature': signature
            }]
        })

    def __repr__(self):
        return f'<{type(self).__name__}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

import pydantic


class ICommand(Protocol):
    __module__: str = 'canonical.lib.protocols'
    # TODO: This basically mirrors an aorta.Command
    # object. Update aorta to have a more sane public
    # interface.
    __envelope__: type[pydantic.BaseModel]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import cast
from typing import overload
from typing import Any
from typing import Annotated
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeVar

from canonical.ext.api.protocols import IObjectIdentifier
from .apimodelfield import APIModelField
from .apiversioned import APIVersioned
from .bases import BaseResource
from .builders import APIResourceClassBuilder
from .objectmeta import ObjectMeta
from .objectreference import ObjectReference
from .apiversionedmeta import APIVersionedMeta

K = TypeVar('K', bound=str)
M = TypeVar('M')


class APIResource(APIVersioned, BaseResource, Generic[K], abstract=True):
    __abstract__ = True
    __builder_class__ = APIResourceClassBuilder

    metadata: Annotated[
        ObjectMeta[K],
            APIModelField(
            title='Metadata',
            description=(
                "`ObjectMeta` is metadata that all persisted resources "
                "must have, which includes all objects users must create."
            ),
            when={'create', 'store', 'update', 'view'}
        )
    ]

    @property
    def cache_key(self):
        return self.metadata.cache_key(self.__meta__)

    @property
    def key(self) -> IObjectIdentifier[Self]:
        return self.get_object_ref(self.__meta__)\
            .with_model(self.model)

    @property
    def model(self):
        return type(self)

    @property
    def name(self) -> K:
        return self.metadata.name

    @overload
    def get_controller(self, require: Literal[True], model: type[M]) -> IObjectIdentifier[M]:
        ...

    @overload
    def get_controller(self, require: Literal[False], model: type[M]) -> IObjectIdentifier[M] | None:
        ...

    def get_controller(self, require: bool, model: type[M]) -> IObjectIdentifier[M] | None:
        ctrl = self.metadata.get_controller()
        if require and not ctrl:
            raise ValueError(
                f"{self.model.__name__} object is expected to have a "
                f"controller, but has none (name: {self.metadata.name}, "
                f"namespace: {self.get_namespace()})."
            )
        if ctrl:
            ctrl.with_model(model)
        return cast(IObjectIdentifier[M], ctrl)

    @overload
    def get_namespace(self, require: Literal[False] = False) -> str | None: ...

    @overload
    def get_namespace(self, require: Literal[True]) -> str: ...

    def get_namespace(self, require: bool = False) -> str | None:
        ns = self.metadata.get_namespace()
        if require and not ns:
            raise TypeError("Object is not namespaced.")
        return ns

    def get_object_ref(self, meta: APIVersionedMeta[Any]) -> ObjectReference:
        ref = ObjectReference(
            api_version=meta.api_version,
            kind=meta.kind,
            name=str(self.name)
        )
        ref.attach(self.metadata)
        return ref

    def in_namespace(self, namespace: str):
        return self.metadata.in_namespace(namespace)

    def replacable(self) -> bool:
        return True

    def update(self, old: Self | None) -> tuple[Self, bool]:
        """Return a tuple containing an instance with updated
        metadata and a boolean indicating if there are any
        changed that need to be persisted.
        """
        if old is None:
            return self, True
        generation = self.metadata.generation
        os = old.model_dump(exclude={'api_version', 'kind', 'metadata', 'status'})
        ns = self.model_dump(exclude={'api_version', 'kind', 'metadata', 'status'})
        if os != ns:
            generation += 1
        data = {
            **self.model_dump(exclude={'metadata'}),
            'metadata': {
                **self.metadata.model_dump(exclude={'generation', 'updated'}),
                'generation': generation,
                'updated': datetime.datetime.now(datetime.timezone.utc),
            }
        }
        return self.model_validate(data), True


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import pydantic

from canonical.lib.types import HTTPResourceLocator


AuthorizationCode: TypeAlias = Annotated[
    str | None,
    pydantic.Field(
        title="Authorization Code",
        description=(
            "The authorization code generated by the authorization "
            "server."
        )
    )
]

Issuer: TypeAlias = Annotated[
    HTTPResourceLocator | None,
    pydantic.Field(
        title="Issuer Identifier",
        description=(
            "The iss parameter value is the issuer identifier of the authorization "
            "server which created the authorization responsed. Its value **must** "
            "be a URL that uses the `https` scheme without any query or fragment "
            "components. If the authorization server provides metadata, the "
            "value of the parameter `iss` **must** be identical to the "
            "authorization server metadata value `issuer`."
        )
    )
]

State: TypeAlias = Annotated[
    str | None,
    pydantic.Field(
        title="State",
        description=(
            "An opaque value used by the client to maintain state between the "
            "request and callback.  The authorization server includes this "
            "value when redirecting the user-agent back to the client.  The "
            "parameter **should** be used for preventing cross-site request "
            "forgery"
        )
    )
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.lib.utils.encoding import b64encode
from canonical.ext.api import ClusterObjectReference
from .bases import BaseClient
from .clientspec import ClientSpec


class ClusterClient(
    BaseClient[ClientSpec[ClusterObjectReference]],
    group='oauth',
    version='v2',
    plural='clusterclients',
    namespaced=False
):

    @property
    def id(self) -> str:
        return b64encode(f'clusterclients::{self.metadata.name}', encoder=str)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from .baserolebinding import BaseRoleBinding


class RoleBinding(
    BaseRoleBinding[Literal['ClusterRole', 'Role']],
    kind='RoleBinding',
    group='iam.webiam.io',
    version='v1',
    plural='rolebindings',
    namespaced=True
):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .grantmodel import GrantModel


class DeviceCodeRequest(GrantModel[Literal['urn:ietf:params:oauth:grant-type:device_code']]):
    device_code: str = pydantic.Field(
        default=...,
        title="Device Code",
        description=(
            "The device verification code, `device_code` "
            "from the device authorization response"
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import json
import operator
import secrets
from typing import Any
from typing import Callable
from typing import Iterable

import pydantic
from canonical import UnixTimestamp
from canonical.ext.jose.types import AudienceType
from canonical.lib.utils.encoding import b64encode_json


class JWT(pydantic.BaseModel):
    _now: datetime.datetime = pydantic.PrivateAttr()

    iss: str | None = pydantic.Field(
        default=None,
        alias='iss',
        title="Issuer",
        description=(
            "The `iss` (issuer) claim identifies the principal "
            "that issued the JWT."
        )
    )

    sub: str | None = pydantic.Field(
        default=None,
        alias='sub',
        title="Subject",
        description=(
            "The `sub` (subject) claim identifies the principal "
            "that is the subject of the JWT."
        )
    )

    aud: AudienceType = pydantic.Field(
        default_factory=AudienceType,
        alias='aud',
        title="Audience",
        description=(
            "The `aud` (audience) claim identifies the recipients "
            "that the JWT is intended for."
        )
    )

    exp: UnixTimestamp | None = pydantic.Field(
        default=None,
        title="Expiration Time",
        description=(
            "The `exp` (expiration time) claim identifies the expiration "
            "time on or after which the JWT **must not** be accepted for "
            "processing. The processing of the `exp` claim requires that "
            "the current date/time **must** be before the expiration "
            "date/time listed in the `exp` claim."
        )
    )

    iat: UnixTimestamp | None = pydantic.Field(
        default=None,
        title="Issued At",
        description=(
            "The `iat` (issued at) claim identifies the time at which "
            "the JWT was issued.  This claim can be used to determine "
            "the age of the JWT."
        )
    )

    nbf: UnixTimestamp | None = pydantic.Field(
        default=None,
        title="Not Before",
        description=(
            "The `nbf` (not before) claim identifies the time before "
            "which the JWT **must not** be accepted for processing.  The "
            "processing of the `nbf` claim requires that the current "
            "date/time **must** be after or equal to the not-before "
            "date/time listed in the `nbf` claim."
        )
    )

    jti: str | None = pydantic.Field(
        default=None,
        title="JWT ID",
        description=(
            "The `jti` (JWT ID) claim provides a unique identifier for the JWT."
        )
    )

    claims: dict[str, Any] = pydantic.Field(default={}, alias='__claims__')

    @classmethod
    def new(cls, ttl: int, **kwargs: Any):
        """Create a new :class:`JWT`."""
        now = datetime.datetime.now(datetime.timezone.utc)
        return cls.model_validate({
            'jti': secrets.token_urlsafe(48),
            **kwargs,
            'exp': (now + datetime.timedelta(seconds=ttl)),
            'iat': now,
            'nbf': now,
        })

    @pydantic.model_validator(mode='before') # type: ignore
    def preprocess(cls, values: bytes | str | dict[str, Any]) -> dict[str, Any]:
        if isinstance(values, (bytes, str)):
            values = json.loads(values)
        assert isinstance(values, dict)
        return {
            **values,
            '__claims__': {
                k: v
                for k, v in values.items()
                if k not in cls.model_fields
            }
        }

    @pydantic.model_serializer(mode='wrap', when_used='always')
    def serialize_model(
        self,
        info: Callable[['JWT'], dict[str, Any]]
    ):
        values = info(self)
        if 'claims' in values:
            values = {**values.pop('claims'), **values}
        return values

    def model_dump_json(self, **kwargs: Any) -> str:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        return super().model_dump_json(**kwargs)

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        return super().model_dump(**kwargs)

    def encode(self, **kwargs: Any) -> bytes:
        """Encode the payload of the JWT."""
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_none', True)
        return b64encode_json({
            **self.claims,
            **self.model_dump(**kwargs)
        })

    def has_claims(self, claims: Iterable[str]) -> bool:
        """Return a boolean indicating if the token has the given claims."""
        return set(claims) <= set(self.model_dump(exclude_none=True))

    def model_post_init(self, _: Any) -> None:
        self._now = datetime.datetime.now(datetime.timezone.utc)

    def has_audience(self, value: str) -> bool:
        return value in self.aud

    def validate_aud(self, aud: str | set[str]):
        return any([
            isinstance(aud, str) and aud in self.aud,
            isinstance(aud, set) and  bool(aud & self.aud),
        ])

    def validate_exp(self, now: datetime.datetime | None = None):
        return self._validate_timestamp(self.exp, operator.ge, now)

    def validate_iat(
        self,
        max_age: int,
        now: datetime.datetime | None = None,
        required: bool = True
    ):
        if self.iat is None:
            return not required
        exp = (now or self._now) - datetime.timedelta(seconds=max_age)
        return self.iat > exp

    def validate_iss(self, issuer: str) -> bool:
        return all([
            self.iss is not None,
            self.iss == issuer
        ])

    def validate_nbf(
        self,
        now: datetime.datetime | None = None,
        required: bool = True
    ):
        return any([
            self.nbf is None and not required,
            self._validate_timestamp(self.nbf, operator.le, now)
        ])

    def _validate_timestamp(
        self,
        value: datetime.datetime | None,
        op: Callable[[datetime.datetime, datetime.datetime], bool],
        now: datetime.datetime | None = None
    ) -> bool:
        if value is None:
            return False
        return op(value, (now or self._now))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import NotRequired
from typing import TypedDict

from canonical.lib.types import EmailAddress
from .audiencetype import AudienceType


class JWTDict(TypedDict):
    iss: NotRequired[str]
    sub: NotRequired[str]
    aud: NotRequired[AudienceType]
    exp: NotRequired[int]
    nbf: NotRequired[int]
    iat: NotRequired[int]
    jti: NotRequired[str]
    name: NotRequired[str]
    given_name: NotRequired[str]
    family_name: NotRequired[str]
    middle_name: NotRequired[str]
    nickname: NotRequired[str]
    preferred_username: NotRequired[str]
    profile: NotRequired[str]
    picture: NotRequired[str]
    website: NotRequired[str]
    email: NotRequired[EmailAddress]
    email_verified: NotRequired[bool]
    gender: NotRequired[str]
    birthdate: NotRequired[datetime.date | str]
    zoneinfo: NotRequired[str]
    locale: NotRequired[str]
    phone_number: NotRequired[str]
    phone_number_verified: NotRequired[str]
    address: NotRequired[str]
    updated_at: NotRequired[int]
    azp: NotRequired[str]
    nonce: NotRequired[str]
    auth_time: NotRequired[str]
    at_hash: NotRequired[str]
    c_hash: NotRequired[str]
    acr: NotRequired[str]
    amr: NotRequired[list[str]]
    #sub_jwk: NotRequired[str]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Iterable

from canonical.utils import class_property
from canonical.utils import merge_signatures
from .protocols import IResourceAuthorizationContext
from .types import PermissionSet


class BaseAuthorizationContext(IResourceAuthorizationContext):

    @class_property
    def __signature__(cls) -> inspect.Signature:
        return merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.require)
        ])

    def __init__(self, **kwargs: Any) -> None:
        self.require(**kwargs)

    def require(self, **_: Any):
        pass

    def get_permission_name(self, api_group: str, plural: str, verb: str) -> str:
        domain = None
        service = api_group
        permission = f'{plural}.{verb}'
        if api_group == '':
            api_group = 'core'
        if api_group and api_group.find('.') != -1:
            service, domain = str.split(api_group, '.', 1)
        permission = f'{service}.{plural}.{verb}'
        if domain is not None:
            permission = f'{domain}/{service}.{plural}.{verb}'
        return permission

    def has_permissions(self, permissions: Iterable[str]) -> set[str]:
        raise NotImplementedError

    def is_authenticated(self) -> bool:
        return False

    def is_authorized(self) -> bool:
        return False

    async def has(self, permissions: str | set[str]) -> bool:
        raise NotImplementedError

    async def get_permissions(self, permissions: set[str]) -> PermissionSet:
        raise NotImplementedError

    async def setup(self):
        pass

    async def teardown(self):
        pass

    async def _setup(self):
        await self.setup()
        return self

    def __await__(self):
        return self._setup().__await__()

    async def __aenter__(self):
        await self.setup()
        return self

    async def __aexit__(
        self,
        cls: type[Exception],
        exception: Exception,
        traceback: Any
    ) -> None:
        await self.teardown()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.types import RedirectURI
from .authorizationstatespec import AuthorizationStateSpec


class ObtainingStateSpec(AuthorizationStateSpec[Literal['obtaining']]):

    def get_redirect_uri(self) -> RedirectURI | None:
        raise NotImplementedError

    def on_obtained(self, response: TokenResponse) -> dict[str, Any]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import binascii

from .base64type import Base64Type


__all__: list[str] = [
    'Base64Int'
]


class Base64Int(int, Base64Type[int]):
    endianness: str = 'big'

    @classmethod
    def b64input(cls, value: bytes) -> int:
        return int(binascii.b2a_hex(value), 16)

    @classmethod
    def b64output(cls, value: int) -> bytes:
        return int.to_bytes(value, (value.bit_length() + 7) // 8, 'big')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical.ext.oauth.types import RedirectURI
from .grantmodel import GrantModel


class AuthorizationCodeRequest(GrantModel[Literal['authorization_code']]):
    code: str = pydantic.Field(
        default=...,
        title="Authorization Code",
        description=(
            "The authorization code received from the authorization server."
        )
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None,
        title="Redirect URI",
        description=(
            "Required, if the `redirect_uri` parameter was included in the "
            "authorization request, and their values **must** be identical."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import TypeVar

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


T = TypeVar('T')


class DomainACL:
    __module__: str = 'canonical.security'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        schema = core_schema.union_schema([
            core_schema.chain_schema([
                core_schema.list_schema(),
                core_schema.no_info_plain_validator_function(cls.fromiterable)
            ]),
            core_schema.chain_schema([
                core_schema.none_schema(),
                core_schema.no_info_plain_validator_function(cls.null)
            ])   
        ])
        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                schema,
                core_schema.is_instance_schema(cls)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(lambda x: x.serialize())
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.union_schema([
            core_schema.list_schema(),
            core_schema.none_schema()
        ]))

    @classmethod
    def fromiterable(cls, v: Iterable[str]) -> 'DomainACL':
        return cls(
            allow=[x for x in v if not str.startswith(x, '~')],
            deny=[x[1:] for x in v if str.startswith(x, '~')],
        )

    @classmethod
    def null(cls, v: None = None) -> 'DomainACL':
        return cls(None, None)

    def __init__(self, allow: Iterable[str] | None, deny: Iterable[str] | None) -> None:
        self._allow = None
        self._deny = None
        if allow is not None or deny is not None:
            self._allow = set(allow or [])
            self._deny = set(deny or [])

    def allows(self, domain: str) -> bool:
        return any([
            not self,
            all([
                self._allow is not None and 'ALL' in self._allow,
                any([
                    self._deny is None,
                    self._deny is not None and domain not in self._deny
                ])
            ]),
            all([
                self._allow is not None and domain in self._allow,
                self._deny is not None and domain not in self._deny
            ])
        ])

    def __bool__(self) -> bool:
        return bool(self._allow is not None or self._deny is not None)

    def serialize(self) -> list[str] | None:
        if not self:
            return None
        assert self._allow is not None
        assert self._deny is not None
        return [
            *[x for x in self._allow],
            *[f'~{x}' for x in self._deny]
        ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .responsemodel import ResponseModel


class TokenResponse(ResponseModel):
    access_token: str = pydantic.Field(
        default=...,
        title="Access Token",
        description=(
            "The access token issued by the authorization server."
        )
    )

    token_type: str = pydantic.Field(
        default=...,
        title="Token Type",
        description=(
            "The type of the token issue. Value is case insensitive."
        )
    )

    expires_in: int | None = pydantic.Field(
        default=None,
        title="Time-to-Live",
        description=(
            "The lifetime in seconds of the access token.  For example, the "
            "value `3600` denotes that the access token will expire in one "
            "hour from the time the response was generated."
        )
    )

    refresh_token: str | None = pydantic.Field(
        default=None,
        title="Refresh Token",
        description=(
            "The refresh token, which can be used to obtain new access "
            "tokens using the same authorization grant"
        )
    )

    scope: str | None = pydantic.Field(
        default=None,
        title="Scope",
        description=(
            "Optional, if identical to the scope requested by the client; "
            "otherwise, required."
        )
    )

    state: str | None = pydantic.Field(
        default=None,
        title="State",
        description=(
            "Required if the `state` parameter was present in the client "
            "authorization request. The exact value received from the "
            "client."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

from canonical.ext.api import APIModelField
from canonical.ext.api import TypedLocalObjectReference
from .namespacespec import NamespaceSpec


RefType = TypeVar('RefType', bound=TypedLocalObjectReference)


class HierarchicalNamespaceSpec(NamespaceSpec, Generic[RefType]):
    parent_ref: RefType = APIModelField(
        default=None,
        alias='parent',
        description=(
            "An optional reference to a parent Resource.\n\nSupported parent "
            "types include `organization` and `folder`. Once set, the parent "
            "cannot be cleared. The parent can be set on creation or "
            "using the `projects.update` method; the end user must "
            "have the `projects.create` permission on the parent."
        ),
        when={'create', 'store', 'view'}
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .builder import BaseBuilder
from .reference import BaseReference
from .resourcerepository import BaseResourceRepository
from .resource import BaseResource
from .resourcebuilder import BaseResourceBuilder
from .resourcedto import BaseResourceDTO
from .rootresource import BaseRootResource


__all__: list[str] = [
    'BaseBuilder',
    'BaseReference',
    'BaseResource',
    'BaseResourceBuilder',
    'BaseResourceDTO',
    'BaseResourceRepository',
    'BaseRootResource',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from collections import abc
from typing import cast
from typing import Any

import pydantic


class FormDataMixin:
    model_dump = pydantic.BaseModel.model_dump

    def encode_form(self, name: str, value: Any, encoding: str) -> list[tuple[str, str]]:
        # TODO: Not very elegant, but covers most use cases.
        params: list[tuple[str, str]] = []
        if isinstance(value, abc.Mapping):
            raise ValueError("Can not dump dictionary to url encoded.")
        if isinstance(value, (str, bytes)):
            if isinstance(value, bytes):
                value = bytes.decode(value, encoding)
            params.append((name, value))
        elif isinstance(value, abc.Collection):
            value = cast(list[Any], value)
            assert isinstance(value, (list, set)), repr(value)
            for item in cast(list[Any], value):
                params.extend(self.encode_form(name, item, encoding))
        else:
            params.append((name, str(value)))
        return params

    def model_dump_urlencoded(self, encoding: str = 'utf-8', mode: str | None = None):
        if mode not in {None, 'query'}:
            raise NotImplementedError
        params: list[tuple[str, str]] = []
        data: dict[str, Any] = self.model_dump(mode='json', exclude_none=True)
        for name, value in data.items():
            params.extend(self.encode_form(name, value, encoding))
        return urllib.parse.urlencode(params, doseq=True)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import httpx
import pydantic

from canonical.lib import FormDataModel
from canonical.ext.oauth.types import Error
from canonical.ext.oauth.types import ProtocolViolation
from canonical.ext.oauth.models.responses import ResponseModel


T = TypeVar('T', bound=ResponseModel)
E = Error | ProtocolViolation


class RequestModel(FormDataModel):
    model_config = {'extra': 'forbid'}

    async def execute(
        self,
        client: httpx.AsyncClient,
        response_model: type[T],
        url: str,
        method: str = 'POST',
    ) -> T | E:
        """Issues a `POST` request using the client to the given
        endpoint `url`.
        """
        adapter: pydantic.TypeAdapter[T | E] = pydantic.TypeAdapter(response_model | E)
        response = await client.request(
            method=method,
            url=url,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/x-www-form-urlencoded'
            },
            content=self.model_dump_urlencoded()
        )
        return adapter.validate_json(response.content)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import APIResourceType
    from .meta import APIVersionedMeta


__all__: list[str] = [
    'get_meta',
    'register',
]


class ResourceRegistry:
    models: dict[str, APIVersionedMeta[APIResourceType]]

    def __init__(self) -> None:
        self.models = {}

    def register(self, meta: APIVersionedMeta[APIResourceType]) -> None:
        if meta.plural:
            self.models[meta.plural] = meta
        if meta.kind:
            self.models[meta.kind] = meta

    def get_meta(
        self,
        api_group: str, *,
        kind: str | None = None,
        plural: str | None = None
    ) -> APIVersionedMeta[APIResourceType]:
        if not bool(kind) ^ bool(plural):
            raise TypeError(
                "Specify either the 'kind' or 'plural' parameter, "
                "but not both."
            )
        return self.models[str(kind or plural)]


_registry: ResourceRegistry = ResourceRegistry()
get_meta = _registry.get_meta
register = _registry.register


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Generic
from typing import NotRequired
from typing import TypeVar
from typing import TypedDict

from canonical.lib.protocols import IBuilder
from .objectmeta import ObjectMeta
from .ownerreference import OwnerReference
from .rootresource import ResourceType


T = TypeVar('T')


class APIResourceBuilder(IBuilder[T], Generic[T]):
    block_owner_deletion: bool = True
    metadata: 'ObjectMetadataDict'
    needs_controller: bool = True

    class ObjectMetadataDict(TypedDict):
        annotations: dict[str, str]
        created: datetime.datetime
        labels: dict[str, str]
        namespace: NotRequired[str]
        owner_references: list[OwnerReference]
        tags: list[str]
        updated: datetime.datetime

    def __init__(self, initial: dict[str, Any] | None = None):
        self.now = datetime.datetime.now(datetime.timezone.utc)
        self.metadata = {
            'annotations': {},
            'created': self.now,
            'labels': {},
            'owner_references': [],
            'tags': [],
            'updated': self.now,
        }
        self.spec = initial or {}

    def annotate(self, name: str, value: str):
        self.metadata['labels'][name] = value

    def annotations(self, annotations: dict[str, str]):
        self.metadata['annotations'].update(annotations)

    def has_controller(self) -> bool:
        return any([x.controller for x in self.metadata['owner_references']])

    def label(self, name: str, value: str):
        self.metadata['labels'][name] = value

    def labels(self, labels: dict[str, str]):
        self.metadata['labels'].update(labels)

    def own(self, owner: ResourceType, controller: bool = False):
        assert owner.metadata.uid
        if controller and owner.is_namespaced():
            ns = self.metadata.get('namespace')
            if ns and ns != owner.metadata.namespace:
                raise ValueError(
                    f"Owner must be in the same namespace (namespace={ns})."
                )
            self.metadata['namespace'] = owner.metadata.namespace
        self.metadata['owner_references'].append(
            OwnerReference.model_validate({
                'api_version': owner.api_version,
                'kind': owner.kind,
                'uid': owner.metadata.uid,
                'name': owner.metadata.name,
                'controller': controller,
                'block_owner_deletion': self.block_owner_deletion
            })
        )

    async def build(self) -> T:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from typing import cast
from typing import Any
from typing import AsyncIterable
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client

from canonical.protocols import ITransaction
from canonical.protocols import ITyped
from canonical.ext.resource import IResourceRepository
from canonical.ext.resource import OwnerReference
from canonical.ext.resource import PrimaryKey
from canonical.ext.resource import ResourceStatus
from canonical.ext.resource import ResourceType
from canonical.ext.resource import ResourceTypeVar
from canonical.ext.resource import StatefulResource
from .basedatastorestorage import BaseDatastoreStorage
from .protocols import IDatastoreKey
from .protocols import IDatastoreEntity
from .protocols import IDatastoreTransaction


T = TypeVar('T')
M = TypeVar('M')


class ResourceRepository(BaseDatastoreStorage, IResourceRepository[IDatastoreTransaction]):
    cluster_namespace: str
    logger: logging.Logger = logging.getLogger('uvicorn')

    async def allocate(self, obj: type[ResourceType]) -> int:
        k = f'{obj.group}/{obj.__name__}'
        i = await self.allocate_identifier(k)
        self.logger.info("Allocated identifier (kind: %s, uid: %s)", k, i)
        return i

    def all(
        self,
        model: type[ResourceTypeVar],
        namespace: str | None = None
    ) -> AsyncIterable[ResourceTypeVar]:
        return self.query( # type: ignore
            model=model,
            namespace=namespace
        )

    def get_entity_name(self, cls: type[ResourceType]) -> str:
        name = f'{cls.group}/{cls.__name__}'
        if not cls.group:
            name = cls.__name__
        return name

    def resource_key(self, resource: ResourceType, model: type[ResourceType] | None = None) -> IDatastoreKey:
        args: list[Any] = [self.get_entity_name(model or type(resource)), resource.metadata.name]
        return self.client.key(*args, namespace=resource.get_namespace()) # type: ignore

    async def exists(self, key: ITyped[ResourceType] | OwnerReference) -> bool:
        if not isinstance(key, (OwnerReference, PrimaryKey)):
            raise NotImplementedError
        filters: list[tuple[str, str, Any]] = [
            ('metadata.name', '=', key.name)
        ]
        if isinstance(key, PrimaryKey):
            q = self.query(model=key.model, filters=filters, namespace=key.namespace)
        else:
            raise NotImplementedError
        return await q.exists()

    async def delete(
        self,
        object: M,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> M:
        raise NotImplementedError

    async def first(
        self,
        model: type[T],
        sort: list[str] | None = None,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> T | None:
        raise NotImplementedError

    async def get(
        self,
        key: ITyped[T],
        *,
        model: type[T] | None = None,
        cached: bool = False,
        max_age: int = 0,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> T:
        k = cast(PrimaryKey[pydantic.BaseModel], key) # type: ignore
        if cached:
            raise NotImplementedError
        obj = await self.get_model_by_key(
            cls=cast(type[pydantic.BaseModel], model) or k.get_type(),
            pk=k.name,
            namespace=k.get_namespace()
        )
        if obj is None:
            raise self.DoesNotExist
        return cast(T, obj)

    async def get_by_name(
        self,
        model: type[ResourceType],
        name: str | int,
        namespace: str | None = None
    ) -> ResourceType:
        obj = await self.get_model_by_key(model, name, namespace=namespace)
        if obj is None:
            raise self.DoesNotExist
        return obj

    async def persist(
        self,
        object: ResourceType,
        transaction: ITransaction | None = None,
        model: type[ResourceType] | None = None,
        **kwargs: Any
    ) -> ResourceType:
        async with self.transaction(transaction) as tx:
            exclude = {}
            if object.has_state():
                exclude = {
                    'status': {'conditions'}
                }
            key = self.resource_key(object, model=model)
            entity = self.entity_factory(
                key=key,
                obj=object,
                exclude_fields=exclude
            )
            await self.persist_entity(tx or self.client, entity)
            if object.has_state():
                await self.persist_status(cast(StatefulResource[Any, Any], object), key, tx)
        return object

    async def persist_entity(
        self,
        client: Client | ITransaction,
        entity: IDatastoreEntity
    ) -> IDatastoreEntity:
        return await self.run_in_executor(functools.partial(client.put, entity)) # type: ignore

    async def persist_status(
        self,
        object: StatefulResource[Any, ResourceStatus],
        key: IDatastoreKey,
        transaction: ITransaction
    ) -> None:
        if not object.status:
            return
        assert object.group
        for condition in object.status.conditions:
            if not condition.is_dirty():
                continue
            assert condition.uid is None
            dao = object.status.storage_model({
                **condition.model_dump(),
                'observed_generation': object.metadata.generation,
                'name': object.metadata.name,
                'namespace': object.metadata.get_namespace(),
                'parent': object.metadata.uid,
                'uid': await self.allocate_identifier('Condition')
            })

            # TODO: ensure that there are no other components writing
            # to the namespace.
            assert object.group
            entity = self.entity_factory(
                key=self.entity_key(
                    kind=f'{type(object).__name__}Condition',
                    identifier=dao.uid,
                    namespace=object.group
                ),
                obj=dao,
            )
            entity['parent'] = key
            await self.persist_entity(transaction, entity)

        # Also persist the status in the service namespace with
        # some additional metadata.
        status = object.clusterstate()

        assert isinstance(status.uid, str) or status.uid > 0
        entity = self.entity_factory(
            key=self.entity_key(
                kind=type(status).__name__,
                identifier=status.uid,
                namespace=object.group
            ),
            obj=status,
        )
        await self.persist_entity(transaction, entity)

    async def pop(
        self,
        key: ITyped[M] | int | str,
        model: type[M] | None = None,
        sort: list[str] | None = None,
        transaction: IDatastoreTransaction | None = None,
        **kwargs: Any
    ) -> M:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable

import fastapi
from canonical.protocols import ITemplateService
from jinja2 import BaseLoader
from jinja2 import ChoiceLoader
from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import PackageLoader


class TemplateService(ITemplateService):
    default_extensions: list[str] = [
        'jinja2.ext.i18n',
        'jinja_markdown.MarkdownExtension'
    ]
    default_loaders: list[BaseLoader] = [
        PackageLoader('canonical.ext.templates')
    ]
    default_dirs: list[str] = []

    def __init__(
        self,
        env: Environment | None = None,
        request: fastapi.Request | None = None,
        template_dirs: Iterable[str] | None = None,
        packages: Iterable[str] | None = None
    ) -> None:
        self.env = env or Environment(
            extensions=self.default_extensions,
            loader=ChoiceLoader([
                FileSystemLoader([*(template_dirs or []), *self.default_dirs]),
                *[PackageLoader(x) for x in packages or []],
                *self.default_loaders,
            ])
        )
        if env is None:
            self.env.install_null_translations() # type: ignore
        self.request = request

    def clone(
        self,
        request: fastapi.Request,
        template_dirs: Iterable[str] | None,
        packages: Iterable[str] | None,
    ) -> ITemplateService:
        return type(self)(
            request=request,
            template_dirs=template_dirs,
            packages=packages
        )

    async def get_template(self, template_name: str, using: str | None = None):
        return self.env.get_template(template_name)

    async def render_template(
        self,
        templates: list[str] | str,
        context: dict[str, Any]
    ) -> str:
        if not isinstance(templates, list):
            templates = [templates]
        t = self.env.select_template(templates)
        if self.request is not None:
            context.setdefault('request', self.request)
        return t.render(**context)
    
    async def select_template(self, templates: list[str], using: str | None = None):
        return self.env.select_template(templates)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing import Iterable

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler
from starlette.routing import compile_path # type: ignore
from starlette.routing import Convertor



class EndpointPath:

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=cls.__default_schema__(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    cls.__default_schema__(),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(cls.__default_schema__())

    @classmethod
    def __default_schema__(cls):
        return core_schema.str_schema(strip_whitespace=True)

    @classmethod
    def validate(cls, v: str):
        if str.startswith(v, '/'):
            raise ValueError(f'{cls.__name__} can not start with a slash: {v}.')
        return cls(*compile_path(v)) # type: ignore

    def __init__(
        self,
        pattern: re.Pattern[str],
        spec: str,
        params: dict[str, Convertor[Any]]
    ):
        self.pattern = pattern
        self.spec = spec
        self.params = params
        self.format_string = re.sub(r'\:{a-z}\}', '', self.spec)

    def construct(self, params: dict[str, Any]):
        if set(params.keys()) != set(self.params.keys()):
            raise TypeError("Missing parameters.")
        return self.format_string.format(**params)

    def match(self, path: str) -> None | dict[str, Any]:
        m = self.pattern.match(path)
        if m is None:
            return None
        params = m.groupdict()
        for k, v in params.items():
            params[k] = self.params[k].convert(v)
        return params

    def validate_keys(self, keys: Iterable[str]) -> bool:
        return set(self.params.keys()) == set(keys)

    def __str__(self):
        return self.spec

    def __repr__(self):
        return f'<EndpointPath: {self.spec}>'


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import Protocol
from typing import TypeVar


T = TypeVar('T')


class IObjectIdentifier(Protocol, Generic[T]):

    def get_model(self) -> type[T]: ...
    def get_namespace(self) -> str | None: ...
    def has_model(self) -> bool: ...
    def with_model(self, model: T) -> 'IObjectIdentifier[T]': ...


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from fastapi.security import HTTPBasic
from fastapi.security import HTTPBasicCredentials

from canonical.ext.oauth.models import ClientAssertion
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.ext.oauth.security import RFC7521
from .tokenrequest import TokenRequest


basic = HTTPBasic(
    auto_error=False,
    description=(
        "Clients in possession of a client password **may** use the HTTP Basic "
        "authentication scheme as defined in RFC 2617 to authenticate with the "
        "authorization server.\n\n"
        "If the client registered a JSON Web Key Set (JWKS) holding it's "
        "public keys, then it **should** use the RFC 7521 Client Assertion "
        "method using a JWT client assertion, even if it also has a client "
        "secret configured."
    ),
    scheme_name='Client Secret (HTTP Basic Auth)'
)

client_assertion = RFC7521(
    description=(
        "An assertion is a package of information that facilitates "
        "the sharing of identity and security information across "
        "security domains. Through the RFC 7521 Assertion Framework "
        "for OAuth 2.0 Client Authentication and Authorization Grants, "
        "clients can employ various mechanisms to authenticate them"
        "selves.\n\nThis server supports the following methods:"
    ),
    scheme_name='RFC 7521 Client Assertion'
)


async def get(
    params: TokenRequest,
    credentials: HTTPBasicCredentials | None = fastapi.Depends(basic),
    assertion: ClientAssertion | None = fastapi.Depends(client_assertion),
) -> str | None:
    if not any([params.client_id, credentials, assertion]):
        # TODO
        raise NotImplementedError
    return 'foo'


TokenEndpointClientDepends = fastapi.Depends(get)
TokenEndpointClient: TypeAlias = Annotated[IRegisteredClient, TokenEndpointClientDepends]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal


TokenTypeLiteral = Literal[
    'urn:ietf:params:oauth:token-type:access_token',
    'urn:ietf:params:oauth:token-type:refresh_token',
    'urn:ietf:params:oauth:token-type:id_token',
    'urn:ietf:params:oauth:token-type:saml1',
    'urn:ietf:params:oauth:token-type:saml2',
    'urn:ietf:params:oauth:token-type:jwt'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import Literal
from typing import TypeVar

import pydantic


K = TypeVar('K')


class RoleReference(pydantic.BaseModel, Generic[K]):
    api_group: Literal['iam.webiam.io'] = pydantic.Field(
        default='iam.webiam.io',
        alias='apiGroup',
        description=(
            "Must be `iam.webiam.io` for `User`, `ServiceAccount` and "
            "`Group` references."
        )
    )

    kind: K = pydantic.Field(
        default=...,
        description=(
            "Specifies the type of resource being referenced."
        )
    )

    name: str = pydantic.Field(
        default=...,
        description=(
            "Identifies the resource being referenced. For reference "
            "to a `Role`, this implies that the resource is in the "
            "same namespace."
        )
    )

    def is_global(self) -> bool:
        return self.kind == 'ClusterRole'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

from .authorizationrequest import AuthorizationRequest
from .authorizationcoderequest import AuthorizationCodeRequest
from .clientcredentialsrequest import ClientCredentialsRequest
from .devicecoderequest import DeviceCodeRequest
from .grantmodel import GrantModel
from .jwtbearerrequest import JWTBearerRequest
from .refreshtokenrequest import RefreshTokenRequest
from .passwordcredentialsrequest import PasswordCredentialsRequest
from .tokenexchangerequest import TokenExchangeRequest


__all__: list[str] = [
    'AuthorizationRequest',
    'AuthorizationCodeRequest',
    'ClientCredentialsRequest',
    'DeviceCodeRequest',
    'GrantModel',
    'JWTBearerRequest',
    'PasswordCredentialsRequest',
    'RefreshTokenRequest',
    'TokenExchangeRequest',
]


TokenRequestType = Union[
    AuthorizationCodeRequest,
    ClientCredentialsRequest,
    DeviceCodeRequest,
    JWTBearerRequest,
    PasswordCredentialsRequest,
    RefreshTokenRequest,
    TokenExchangeRequest
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import importlib
import inspect
from types import ModuleType
from typing import get_args
from typing import get_origin
from typing import Any
from typing import Iterable
from typing import Union

import fastapi

from canonical.ext.api import APIResourceType
from canonical.ext.api import APIVersionedMeta
from canonical.ext.api import APIModelInspector
from canonical.ext.cache import NullCache
from canonical.ext.resource import APIDescriptor
from canonical.ext.resource import APIGroupVersionList
from canonical.ext.resource import APIResourceList
from canonical.ext.resource import Error
from canonical.ext.resource import Resource
from canonical.ext.resource import ResourceServerList
from canonical.ext.resource import RootResource
from canonical.ext.iam.protocols import IResourceAuthorizationContext
from canonical.lib.protocols import ICache
from canonical.utils.http import MediaTypeSelector

from .apiroute import APIRoute
from .params import NegotiateResponseMediaType
from .resourceclient import ResourceClient
from .resourcerouter import ResourceRouter
from .response import Response


MediaType = NegotiateResponseMediaType({'application/yaml', 'application/json', "text/html"})


class ResourceApplication(fastapi.FastAPI):
    inspector: APIModelInspector = APIModelInspector()
    resource_paths: set[str] = set()
    media_types: MediaTypeSelector = MediaTypeSelector({
        'text/html',
        'application/json',
        'application/yaml'
    })

    def __init__(
        self,
        discovers: list[str] | None = None,
        *args: Any, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self.router.route_class = APIRoute
        self.discovers = set(discovers or [])
        self.groups: dict[str, APIGroupVersionList] = {}
        self.resources: dict[str, dict[str, APIResourceList]] = {}
        self.add_api_route(
            methods=['GET'],
            path=f'/',
            endpoint=self.root,
            summary=f"Service descriptor",
            response_description=f"APIDescriptor object.",
            response_model=APIDescriptor,
            include_in_schema=True,
            tags=['Discovery Endpoints']
        )
        if self.discovers:
            self.add_api_route(
                methods=['GET'],
                path=f'/discover',
                endpoint=self.resource_servers,
                summary=f"Resource servers",
                response_description=f"ResourceServerList object.",
                response_model=ResourceServerList,
                include_in_schema=True,
                tags=['Discovery Endpoints']
            )

        @self.exception_handler(404)
        def _(request: fastapi.Request, _: Any) -> Any:
            assert request.client is not None
            return Response(
                media_type=self.media_types.select(request.headers.get('Accept')),
                status_code=404,
                content=Error.factory({
                    'status_code': 404,
                    'detail': "The server cannot find the requested resource",
                    'request': {
                        'url': str(request.url),
                        'host': request.client.host
                    }
                })
            )

    def add(
        self,
        impl: type[Resource[Any] | RootResource[Any] | ResourceRouter[Any]] | ModuleType | str,
        verbs: Iterable[str] | None = None,
        authorization: type[IResourceAuthorizationContext] | None = None,
        cache: ICache = NullCache(),
        **kwargs: Any
    ) -> None:
        root = fastapi.APIRouter(**kwargs)
        self.openapi_tags = self.openapi_tags or []
        verbs = set(verbs or [])
        if isinstance(impl, str):
            impl = importlib.import_module(f'{impl}.routers')
        if inspect.isclass(impl) and issubclass(impl, ResourceRouter):
            router: ResourceRouter[Any] = impl(enabled_verbs=verbs)
            router.add_to_application(
                app=root,
                register=self.register_resource,
                tags=self.openapi_tags,
                authorization=authorization,
                cache=cache
            )
        elif isinstance(impl, ModuleType):
            for _, value in inspect.getmembers(impl):
                if isinstance(value, ResourceRouter):
                    value.add_to_application(
                        app=root,
                        register=self.register_resource,
                        tags=self.openapi_tags,
                        authorization=authorization,
                        cache=cache
                    )
                    continue
                if isinstance(value, fastapi.APIRouter):
                    self.include_router(router=value)
        else:
            raise NotImplementedError
        self.include_router(router=root)

    def create_discovery_endpoint(
        self,
        meta: APIVersionedMeta[Any],
        resources: APIResourceList,
    ) -> None:

        async def discover(media_type: str = MediaType) -> Response[Any]:
            return Response(
                status_code=200,
                media_type=media_type,
                content=self.resources[meta.api_group][meta.version]
            )

        self.add_api_route(
            methods=['GET'],
            path=f'/{resources.path(meta.api_group)}',
            endpoint=discover,
            summary=f"{meta.api_group or 'default'} {meta.version}",
            response_description=f"The list of supported resources.",
            response_model=APIResourceList,
            include_in_schema=True,
            tags=['Discovery Endpoints']
        )

    def create_group_discovery_endpoint(self, api_group: str) -> None:
        async def discover(media_type: str = MediaType) -> Response[Any]:
            return Response(
                status_code=200,
                media_type=media_type,
                content=self.groups[api_group]
            )

        self.add_api_route(
            methods=['GET'],
            path=f'/{api_group}',
            endpoint=discover,
            summary=f"{api_group or 'core'}",
            response_description=f"The list of supported versions.",
            response_model=APIGroupVersionList,
            include_in_schema=True,
            tags=['Discovery Endpoints']
        )

    def openapi(self) -> dict[str, Any]:
        res = super().openapi()
        for _, method_item in res.get("paths", {}).items():
            for method, param in method_item.items():
                responses = param.get("responses")
                # remove default 422 - the default 422 schema is HTTPValidationError
                if "422" in responses and responses["422"]["content"][
                    "application/json"
                ]["schema"]["$ref"].endswith("HTTPValidationError"):
                    del responses["422"]
                for status_code, response in responses.items():
                    status_code = str(status_code)
                    if method == 'head' or status_code.startswith('3'):
                        response.pop('content', None)
                        continue

        return res

    def register_resource(self, model: type[APIResourceType], verb: str, plural: str | None = None) -> None:
        meta = self.inspector.inspect(model)
        if meta.api_group not in self.resources:
            self.resources[meta.api_group] = {}
            self.groups[meta.api_group] = APIGroupVersionList(group=meta.api_group)
            self.create_group_discovery_endpoint(meta.api_group)
        if meta.version not in self.resources[meta.api_group]:
            resources = self.resources[meta.api_group][meta.version] = APIResourceList(
                groupVersion=meta.version
            )
            self.groups[meta.api_group].add(resources)
            self.create_discovery_endpoint(meta, resources)
        resource = self.resources[meta.api_group][meta.version].add(meta, plural=plural)
        resource.add(verb)
        for root in self.inspector.get_root_types(model):
            self.register_resource(root, verb, plural=meta.plural)
            

    async def resource_servers(
        self,
        request: fastapi.Request,
        media_type: str = MediaType
    ) -> Response[ResourceServerList]: # type: ignore
        discovers = [f'{request.url.scheme}://{request.url.netloc}', *self.discovers]
        futures: list[asyncio.Future[Any]] = []
        async with ResourceClient(headers={'Accept': 'application/json'}) as client:
            for server in discovers:
                futures.append(
                    asyncio.ensure_future(client.discover_server(server))
                )
            return Response(
                status_code=200,
                media_type=media_type,
                content=ResourceServerList(
                    servers=await asyncio.gather(*futures)
                )
            )

    async def root(
        self,
        request: fastapi.Request,
        media_type: str = MediaType
    ) -> Response[APIDescriptor]: # type: ignore
        return Response(
            status_code=200,
            media_type=media_type,
            content=APIDescriptor(
                host=request.url.netloc,
                groups=list(self.groups.values())
            )
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing_extensions import SupportsIndex

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


class DomainName(str):
    __module__: str = 'canonical.lib.types'
    pattern: re.Pattern[str] = re.compile(r'^([0-9a-z\-_]+)$')
    min_length: int = 3
    max_length: int = 255

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=cls.__default_schema__(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    cls.__default_schema__(),
                    core_schema.no_info_plain_validator_function(cls.validate)
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(cls.__default_schema__())

    @classmethod
    def __default_schema__(cls):
        return core_schema.str_schema(
            min_length=cls.min_length,
            max_length=cls.max_length
        )

    @classmethod
    def validate(cls, v: Any, _: Any = None) -> str:
        if not isinstance(v, str):
            raise ValueError(f"Can not create DomainName from {type(v).__name__}")
        if not v:
            raise ValueError("A domain name can not be an empty string.")
        if len(v) > 253:
            raise ValueError("Value is too long to be a valid domain name.")
        v = str.lower(v)
        labels: list[str] = str.split(v, '.')
        for label in labels:
            if label.startswith('-') or label.endswith('-'):
                raise ValueError("A DNS label can not start or end with a hyphen.")
            if not cls.pattern.match(label):
                raise ValueError("Invalid characters in DNS label.")
            if len(label) > 63:
                raise ValueError("A DNS label is at most 63 characters.")
        return cls(v)

    def __getitem__(self, __i: SupportsIndex | slice) -> str:
        value = str.split(self, '.')[__i]
        if isinstance(value, list):
            value = str.join('.', value)
        return value


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Callable
from typing import TypeVar
from typing import Protocol

from canonical.utils import Deferred


T = TypeVar('T')


class IDeferrable(Protocol):
    __module__: str = 'canonical.protocols'

    @classmethod
    def defer(cls: T, resolve: Callable[[], T]) -> T:
        return cast(T, Deferred(resolve))


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
import pydantic

from canonical.ext.api import APIModel



class ChildModel(APIModel):
    taz: int


class Model(APIModel):
    model_config = {'extra': 'forbid'}

    foo: int = pydantic.Field(...)
    bar: int = pydantic.Field(..., frozen=True)
    baz: ChildModel | None


class PolyModel(APIModel):
    foo: int


class InvalidModel(APIModel):
    foo: str


def test_replace():
    ours = Model(foo=1, bar=1, baz=ChildModel(taz=3))
    theirs = Model(foo=2, bar=1, baz=ChildModel(taz=2))
    ours.replace(theirs)
    assert ours.foo == theirs.foo
    assert ours.baz == theirs.baz


def test_replace_poly():
    ours = Model(foo=1, bar=1, baz=ChildModel(taz=3))
    theirs = PolyModel(foo=2)
    ours.replace(theirs)
    assert ours.foo == theirs.foo


def test_replace_invalid():
    ours = Model(foo=1, bar=1, baz=ChildModel(taz=3))
    theirs = InvalidModel(foo='aaa')
    with pytest.raises(pydantic.ValidationError):
        ours.replace(theirs)


def test_can_not_replace_frozen_fields():
    ours = Model(foo=1, bar=1, baz=None)
    theirs = Model(foo=2, bar=2, baz=None)
    ours.replace(theirs)
    assert ours.foo == theirs.foo
    assert ours.bar != theirs.bar



# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Self
from typing import TypeVar

import pydantic
import pydantic.fields

from canonical.ext.api.protocols import IObjectIdentifier


T = TypeVar('T')


class BaseReference:
    _default_reference_model: ClassVar[type[Any] | None] = None
    _reference_model: type[Any] | None = pydantic.PrivateAttr(default_factory=lambda: None)

    @classmethod
    def __init_subclass__(cls, ref: type | None = None):
        if ref is not None:
            cls._default_reference_model = ref

    @property
    def scalar(self) -> int | str:
        raise NotImplementedError

    def attach_to_namespace(self, namespace: str) -> Self:
        raise NotImplementedError

    def as_name(self) -> int | str:
        raise NotImplementedError

    def get_model(self) -> type[Any]:
        model = self._reference_model or self._default_reference_model
        if model is None or isinstance(model, pydantic.fields.ModelPrivateAttr):
            raise TypeError(f"{type(self).__name__} is not attached to a model.")
        return model

    def get_namespace(self) -> str | None:
        raise NotImplementedError(
            f'{type(self).__name__}.get_namespace() is not implemented.'
        )

    def has_model(self) -> bool:
        return self._reference_model is not None

    def in_namespace(self, namespace: str) -> bool:
        raise NotImplementedError

    def is_cluster(self) -> bool:
        raise NotImplementedError

    def is_local(self) -> bool:
        raise NotImplementedError

    def is_namespaced(self) -> bool:
        raise NotImplementedError

    def with_model(self, model: type[T]) -> IObjectIdentifier[T]:
        self._reference_model = model
        return self

    def with_namespace(self, namespace: str) -> Self:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


class HTTPRequestModelMixin:
    _method: str | None = None
    _url: str | None = None
    #method: str = pydantic.Field(
    #    default=None,
    #    exclude=True
    #)

    #url: str | None = pydantic.Field(
    #    default=None,
    #    exclude=True
    #)

    def with_endpoint(self, method: str, endpoint: str):
        self._method = method
        self._url = endpoint
        return self


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .permission import Permission
from .permissionwildcard import PermissionWildcard


PermissionType = Permission | PermissionWildcard


class PermissionSet(set[PermissionType]):

    def has(self, permission: str) -> bool:
        for granted in self:
            is_granted = granted.matches(permission)
            if is_granted:
                break
        else:
            is_granted = False
        return is_granted


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any

import pydantic

from .apimodel import APIModel
from .objectmeta import ObjectMeta


class ResourceStatus(APIModel):
    changed: datetime.datetime = pydantic.Field(
        default=...,
        description=(
            "The date and time of the last change to this {kind}."
        )
    )

    version: int = pydantic.Field(
        default=...,
        description=(
            "Current version of the {kind}."
        )
    )

    message: str = pydantic.Field(
        default=...,
        description="The message of the last known condition."
    )

    current: str = pydantic.Field(
        default=...,
        description="The status of the last known condition."
    )

    def update(
        self,
        metadata: ObjectMeta[Any],
        status: str,
        **kwargs: Any
    ) -> None:
        self.changed = datetime.datetime.now(datetime.timezone.utc)
        self.current = status
        self.version = metadata.generation
        for attname, value in kwargs.items():
            setattr(self, attname, value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorize import AuthorizeEndpoint
from .clientredirection import ClientRedirectionEndpoint
from .metadata import MetadataEndpoint
from .token import TokenEndpoint
from .upstreamauthorization import UpstreamAuthorizationEndpoint


__all__: list[str] = [
    'AuthorizeEndpoint',
    'ClientRedirectionEndpoint',
    'MetadataEndpoint',
    'TokenEndpoint',
    'UpstreamAuthorizationEndpoint',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union

import pydantic

from .types import JWECompactEncoded
from .types import JWSCompactEncoded
from .models import JWS
from .protocols import ITokenSignatureVerifier


JOSEType = Union[
    JWS,
    JWECompactEncoded,
    JWSCompactEncoded,
]

class JOSE(pydantic.RootModel[JOSEType]):

    @pydantic.field_validator('root', mode='after')
    def decode_compact(cls, value: Any) -> JWSCompactEncoded | JWECompactEncoded:
        if isinstance(value, JWSCompactEncoded):
            value = value.jose(JWS)
        return value

    @property
    def signatures(self):
        return self._get_signatures()

    @property
    def signature(self):
        return self.signatures[0]

    def compact(self):
        return self.root.compact()

    def is_encoded(self) -> bool:
        return isinstance(self.root, (JWECompactEncoded, JWSCompactEncoded))

    def is_jws(self) -> bool:
        return isinstance(self.root, JWS)

    def model_dump_json(self, **kwargs: Any) -> str:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        return super().model_dump_json(**kwargs)

    def model_post_init(self, _: Any) -> None:
        if isinstance(self.root, JWSCompactEncoded):
            self.root = self.root.jose(JWS)

    @JWS.require
    def verify(self, verifier: ITokenSignatureVerifier):
        assert isinstance(self.root, JWS)
        return self.root.verify(verifier)

    @JWS.require
    def _get_signatures(self):
        assert isinstance(self.root, JWS)
        return self.root.signatures

    def __str__(self):
        return str(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

from redis import asyncio as aioredis

from canonical.ext.cache import BaseCache


class RedisCache(BaseCache):
    __module__: str = 'canonical.ext.redis'

    def __init__(
        self,
        url: str,
        keyalg: Literal['sha256'] | None = None,
        prefix: str | None = None,
        **kwargs: Any
    ):
        super().__init__(keyalg, prefix, **kwargs)
        self.client: aioredis.Redis = aioredis.from_url(url) # type: ignore

    async def fetch(self, key: str) -> bytes | None:
        return await self.client.get(key) # type: ignore

    async def put(
        self,
        key: str,
        value: Any,
        encrypt: bool = False,
        ttl: int | None = None
    ) -> None:
        await self.client.set(key, value, ex=ttl) # type: ignore



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.lib.utils.encoding import b64encode
from .jsonwebkey import JSONWebKey


A = TypeVar('A')


class JOSEHeader(pydantic.BaseModel, Generic[A]):
    alg: A
    cty: str | None = None
    crit: list[str] = []
    jku: str | None = None
    jwk: JSONWebKey | None = None
    kid: str | None = None
    typ: str | None = None
    x5c: str | None = None
    x5t: str | None = None
    x5t_sha256: str | None = pydantic.Field(default=None, alias='x5t#S256')
    x5u: str | None = None
    claims: dict[str, Any] = {}

    @pydantic.model_validator(mode='before') # type: ignore
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        return {
            **{
                name: values[name] for name in cls.model_fields.keys()
                if name in values and name != 'claims'
            },
            **(values.get('claims') or {})
        }

    def encode(self, **kwargs: Any) -> bytes:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        exclude = kwargs.setdefault('exclude', set())
        exclude.add('claims')
        claims = self.model_dump(**kwargs)
        claims.update(self.claims)
        return b64encode(json.dumps(claims))

    def __bytes__(self):
        return self.encode()


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import TypeVar

import pydantic

from .builder import BaseBuilder


T = TypeVar('T')


class BaseResource:
    __builder__: ClassVar[BaseBuilder[Any] | None] = None
    __builder_class__: ClassVar[type[BaseBuilder[Any]]]
    __create_model__: ClassVar[type[pydantic.BaseModel]]
    __namespaced__: ClassVar[bool]

    @property
    def cache_key(self) -> str:
        raise NotImplementedError

    def replacable(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .grantmodel import GrantModel


class ClientCredentialsRequest(GrantModel[Literal['client_credentials']]):
    scope: str | None = pydantic.Field(
        default=None,
        title="Scope",
        description=(
            "The scope of the access request."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import TypeVar
from typing import Union
from typing import TYPE_CHECKING

import pydantic
from pydantic.fields import FieldInfo

from .apimodelinspector import APIModelInspector
if TYPE_CHECKING:
    from .apimodel import APIModel
    from .apiversioned import APIVersioned
    from .resource import Resource
    from .rootresource import RootResource


T = TypeVar('T', bound=Union['APIModel', 'Resource[Any]', 'RootResource[Any]'])


class APIVersionedMeta(pydantic.BaseModel, Generic[T]):
    inspector: ClassVar[APIModelInspector] = APIModelInspector()
    _model: type[T] | None = pydantic.PrivateAttr(default=None)
    api_version: str
    api_group: str
    base_path: str
    version: str
    kind: str
    namespaced: bool
    plural: str

    @classmethod
    def fromqualname(
        cls,
        model: type[APIVersioned],
        api_version: str,
        **kwargs: Any
    ):
        name, version = str.split(api_version, '/')
        if '.' in name:
            plural, group = name.split('.', 1)
        else:
            group = ''
            plural = name
        base_path = group
        if base_path:
            base_path += '/'
        base_path = f'{base_path}{version}'
        if cls.inspector.is_namespaced(model):
            base_path += '/namespaces/{namespace}'
        cls.base_path = f'{base_path}/{plural}'
        return cls(
            api_version=api_version,
            api_group=group,
            base_path=base_path,
            version=version,
            kind=model.__name__,
            namespaced=cls.inspector.is_namespaced(model),
            plural=plural
        )

    @property
    def model(self) -> type[T]:
        assert self._model
        return self.model

    def contribute_to_class(
        self,
        cls: type[T],
        fields: dict[str, FieldInfo],
        root: tuple[type[Any], ...] | None = None
    ):
        setattr(cls, '__meta__', self)
        if fields:
            fields['api_version'].annotation = Literal[f'{self.api_version}']
            fields['api_version'].default = self.api_version
            fields['kind'].annotation = Literal[f'{self.kind}']
            fields['kind'].default = self.kind